# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
from pathlib import Path
from typing import Collection, List, Optional, Type

import libcst as cst
from libcst.metadata import MetadataWrapper
from libcst.testing.utils import data_provider, UnitTest

from fixit.common.base import CstContext, CstLintRule, LintConfig
from fixit.common.comments import CommentInfo
from fixit.common.ignores import IgnoreInfo
from fixit.common.line_mapping import LineMappingInfo
from fixit.common.report import CstLintRuleReport
from fixit.common.unused_suppressions import (
    _compose_new_comment,
    RemoveUnusedSuppressionsRule,
    UNUSED_SUPPRESSION_CODES_IN_COMMENT_MESSAGE,
    UNUSED_SUPPRESSION_COMMENT_MESSAGE,
)
from fixit.common.utils import dedent_with_lstrip
from fixit.rule_lint_engine import _get_tokens, _visit_cst_rules_with_context


class UsedRule(CstLintRule):
    pass


class UsedRule2(CstLintRule):
    pass


FILE_PATH: Path = Path("fake/path.py")


class RemoveUnusedSuppressionsRuleTest(UnitTest):
    @data_provider(
        {
            "used_suppression_one_code_oneline": {
                "source": dedent_with_lstrip(
                    """
                    # lint-ignore: UsedRule
                    foo = bar
                    """
                ).encode(),
                "rules_in_lint_run": [UsedRule],
                "rules_without_report": [],
                "suppressed_line": 2,
                "expected_unused_suppressions_report_messages": [],
            },
            "used_suppression_one_code_oneline_with_reason": {
                "source": dedent_with_lstrip(
                    """
                    # lint-ignore: UsedRule: reason blah.
                    foo = bar
                    """
                ).encode(),
                "rules_in_lint_run": [UsedRule],
                "rules_without_report": [],
                "suppressed_line": 2,
                "expected_unused_suppressions_report_messages": [],
            },
            "used_suppression_one_code_multiline": {
                "source": dedent_with_lstrip(
                    """
                    # lint-ignore: UsedRule: reason
                    # lint: reason continued blah
                    # lint: blah blah.
                    foo = bar
                    """
                ).encode(),
                "rules_in_lint_run": [UsedRule],
                "rules_without_report": [],
                "suppressed_line": 4,
                "expected_unused_suppressions_report_messages": [],
            },
            "used_suppression_many_codes_oneline": {
                "source": dedent_with_lstrip(
                    """
                    # lint-ignore: UsedRule, UsedRule2
                    foo = bar
                    """
                ).encode(),
                "rules_in_lint_run": [UsedRule, UsedRule2],
                "rules_without_report": [],
                "suppressed_line": 2,
                "expected_unused_suppressions_report_messages": [],
            },
            "used_suppression_many_codes_multiline": {
                "source": dedent_with_lstrip(
                    """
                    # lint-ignore: UsedRule, UsedRule2:
                    # lint: reason blah blah.
                    foo = bar
                    """
                ).encode(),
                "rules_in_lint_run": [UsedRule, UsedRule2],
                "rules_without_report": [],
                "suppressed_line": 3,
                "expected_unused_suppressions_report_messages": [],
            },
            "unused_suppression_one_code_oneline": {
                "source": dedent_with_lstrip(
                    """
                    # lint-ignore: UsedRule: reason blah blah.
                    foo = bar
                    """
                ).encode(),
                "rules_in_lint_run": [UsedRule],
                "rules_without_report": [UsedRule],
                "suppressed_line": 2,
                "expected_unused_suppressions_report_messages": [
                    UNUSED_SUPPRESSION_COMMENT_MESSAGE
                ],
                "expected_replacements": [
                    dedent_with_lstrip(
                        """
                    foo = bar
                    """
                    )
                ],
            },
            "unused_suppression_one_code_multiline": {
                "source": dedent_with_lstrip(
                    """
                    # lint-ignore: UsedRule: reason
                    # lint: reason continued.
                    foo = bar
                    """
                ).encode(),
                "rules_in_lint_run": [UsedRule],
                "rules_without_report": [UsedRule],
                "suppressed_line": 3,
                "expected_unused_suppressions_report_messages": [
                    UNUSED_SUPPRESSION_COMMENT_MESSAGE
                ],
                "expected_replacements": [
                    dedent_with_lstrip(
                        """
                    foo = bar
                    """
                    )
                ],
            },
            "unused_suppression_many_codes_oneline": {
                "source": dedent_with_lstrip(
                    """
                    # lint-ignore: UsedRule, UsedRule2: reason
                    foo = bar
                    """
                ).encode(),
                "rules_in_lint_run": [UsedRule, UsedRule2],
                "rules_without_report": [UsedRule],
                "suppressed_line": 2,
                "expected_unused_suppressions_report_messages": [
                    UNUSED_SUPPRESSION_CODES_IN_COMMENT_MESSAGE.format(
                        lint_codes="UsedRule"
                    )
                ],
                "expected_replacements": [
                    dedent_with_lstrip(
                        """
                # lint-ignore: UsedRule2: reason
                foo = bar
                """
                    )
                ],
            },
            "unused_suppression_many_codes_multiline": {
                "source": dedent_with_lstrip(
                    """
                    # lint-ignore: UsedRule, UsedRule2: reason
                    # lint: reason continued.
                    foo = bar
                    """
                ).encode(),
                "rules_in_lint_run": [UsedRule, UsedRule2],
                "rules_without_report": [UsedRule],
                "suppressed_line": 3,
                "expected_unused_suppressions_report_messages": [
                    UNUSED_SUPPRESSION_CODES_IN_COMMENT_MESSAGE.format(
                        lint_codes="UsedRule"
                    )
                ],
                "expected_replacements": [
                    dedent_with_lstrip(
                        """
                # lint-ignore: UsedRule2: reason reason
                # lint: continued.
                foo = bar
                """
                    )
                ],
            },
            "unused_suppression_many_codes_all_unused": {
                "source": dedent_with_lstrip(
                    """
                    # lint-ignore: UsedRule, UsedRule2: reason
                    # lint: reason continued.
                    foo = bar
                    """
                ).encode(),
                "rules_in_lint_run": [UsedRule, UsedRule2],
                "rules_without_report": [UsedRule, UsedRule2],
                "suppressed_line": 3,
                "expected_unused_suppressions_report_messages": [
                    UNUSED_SUPPRESSION_COMMENT_MESSAGE
                ],
                "expected_replacements": [
                    dedent_with_lstrip(
                        """
                        foo = bar
                        """
                    )
                ],
            },
            "multiple_suppressions": {
                "source": dedent_with_lstrip(
                    """
                    # lint-ignore: UsedRule: first reason
                    # lint: first reason continued.
                    # lint-ignore: UsedRule2: second reason
                    # lint: second reason continued.
                    foo = bar
                    """
                ).encode(),
                "rules_in_lint_run": [UsedRule, UsedRule2],
                "rules_without_report": [UsedRule],
                "suppressed_line": 5,
                "expected_unused_suppressions_report_messages": [
                    UNUSED_SUPPRESSION_COMMENT_MESSAGE
                ],
                "expected_replacements": [
                    dedent_with_lstrip(
                        """
                        # lint-ignore: UsedRule2: second reason
                        # lint: second reason continued.
                        foo = bar
                        """
                    )
                ],
            },
            "multiple_unused_suppressions": {
                "source": dedent_with_lstrip(
                    """
                    # lint-ignore: UsedRule: first reason
                    # lint: first reason continued.
                    # lint-ignore: UsedRule2: second reason
                    # lint: second reason continued.
                    foo = bar
                    """
                ).encode(),
                "rules_in_lint_run": [UsedRule, UsedRule2],
                "rules_without_report": [UsedRule, UsedRule2],
                "suppressed_line": 5,
                "expected_unused_suppressions_report_messages": [
                    UNUSED_SUPPRESSION_COMMENT_MESSAGE,
                    UNUSED_SUPPRESSION_COMMENT_MESSAGE,
                ],
                "expected_replacements": [
                    dedent_with_lstrip(
                        """
                        # lint-ignore: UsedRule2: second reason
                        # lint: second reason continued.
                        foo = bar
                        """
                    ),
                    dedent_with_lstrip(
                        """
                        # lint-ignore: UsedRule: first reason
                        # lint: first reason continued.
                        foo = bar
                        """
                    ),
                ],
            },
            "suppressions_with_unlinted_codes_oneline": {
                "source": dedent_with_lstrip(
                    """
                    # lint-ignore: UnusedRule
                    foo = bar
                    """
                ).encode(),
                "rules_in_lint_run": [UsedRule, UsedRule2],
                "rules_without_report": [],
                "suppressed_line": 2,
                "expected_unused_suppressions_report_messages": [],
            },
            "suppressions_with_unlinted_codes_multiline": {
                "source": dedent_with_lstrip(
                    """
                    # lint-ignore: UnusedRule: reason
                    # lint: reason continued
                    foo = bar
                    """
                ).encode(),
                "rules_in_lint_run": [UsedRule, UsedRule2],
                "rules_without_report": [],
                "suppressed_line": 3,
                "expected_unused_suppressions_report_messages": [],
            },
        }
    )
    def test(
        self,
        *,
        source: bytes,
        rules_in_lint_run: Collection[Type[CstLintRule]],
        rules_without_report: Collection[Type[CstLintRule]],
        suppressed_line: int,
        expected_unused_suppressions_report_messages: Collection[str],
        expected_replacements: Optional[List[str]] = None,
    ) -> None:
        reports = [
            CstLintRuleReport(
                file_path=FILE_PATH,
                node=cst.EmptyLine(),
                code=rule.__name__,
                message="message",
                line=suppressed_line,
                column=0,
                module=cst.MetadataWrapper(cst.parse_module(source)),
                module_bytes=source,
            )
            for rule in rules_in_lint_run
            if rule not in rules_without_report
        ]
        tokens = _get_tokens(source)
        ignore_info = IgnoreInfo.compute(
            comment_info=CommentInfo.compute(tokens=tokens),
            line_mapping_info=LineMappingInfo.compute(tokens=tokens),
            use_noqa=False,
        )
        cst_wrapper = MetadataWrapper(cst.parse_module(source), unsafe_skip_copy=True)
        config = LintConfig(
            rule_config={
                RemoveUnusedSuppressionsRule.__name__: {
                    "ignore_info": ignore_info,
                    "rules": rules_in_lint_run,
                }
            }
        )
        unused_suppressions_context = CstContext(cst_wrapper, source, FILE_PATH, config)
        for report in reports:
            ignore_info.should_ignore_report(report)
        _visit_cst_rules_with_context(
            cst_wrapper, [RemoveUnusedSuppressionsRule], unused_suppressions_context
        )

        messages = []
        patches = []
        for report in unused_suppressions_context.reports:
            messages.append(report.message)
            patches.append(report.patch)

        self.assertEqual(messages, expected_unused_suppressions_report_messages)
        if expected_replacements is None:
            self.assertEqual(len(patches), 0)
        else:
            self.assertEqual(len(patches), len(expected_replacements))

            for idx, patch in enumerate(patches):
                replacement = patch.apply(source.decode())
                self.assertEqual(replacement, expected_replacements[idx])

    def test_compose_new_comment_oneline(self) -> None:
        source = dedent_with_lstrip("# lint-fixme: UsedRule, UsedRule2: reason...")
        tokens = _get_tokens(source.encode())
        ignore_info = IgnoreInfo.compute(
            comment_info=CommentInfo.compute(tokens=tokens),
            line_mapping_info=LineMappingInfo.compute(tokens=tokens),
            use_noqa=False,
        )
        local_suppression_comments = (
            ignore_info.local_ignore_info.local_suppression_comments_by_line[1]
        )
        self.assertEqual(len(local_suppression_comments), 1)
        local_suppression_comment = local_suppression_comments[0]

        # First code unneeded.
        unneeded_codes = ["UsedRule"]
        new_comment_lines = _compose_new_comment(
            local_suppression_comment, unneeded_codes, 1
        )
        expected_new_lines = ["# lint-fixme: UsedRule2: reason..."]
        self.assertEqual(new_comment_lines, expected_new_lines)

        # Second code unneeded.
        unneeded_codes = ["UsedRule2"]
        new_comment_lines = _compose_new_comment(
            local_suppression_comment, unneeded_codes, 1
        )
        expected_new_lines = ["# lint-fixme: UsedRule: reason..."]
        self.assertEqual(new_comment_lines, expected_new_lines)

        # Both codes unneded.
        unneeded_codes = ["UsedRule", "UsedRule2"]
        new_comment_lines = _compose_new_comment(
            local_suppression_comment, unneeded_codes, 1
        )
        expected_new_lines = []
        self.assertEqual(new_comment_lines, expected_new_lines)

        # Both codes needed.
        unneeded_codes = []
        new_comment_lines = _compose_new_comment(
            local_suppression_comment, unneeded_codes, 1
        )
        # Should be unchanged.
        expected_new_lines = [source]
        self.assertEqual(new_comment_lines, expected_new_lines)

    def test_compose_new_comment_multiline(self) -> None:
        source = dedent_with_lstrip(
            """
            # lint-fixme: UsedRule, UsedRule2: reason...
            # lint: reason continued
            """
        )
        tokens = _get_tokens(source.encode())
        ignore_info = IgnoreInfo.compute(
            comment_info=CommentInfo.compute(tokens=tokens),
            line_mapping_info=LineMappingInfo.compute(tokens=tokens),
            use_noqa=False,
        )
        local_suppression_comments = (
            ignore_info.local_ignore_info.local_suppression_comments_by_line[1]
        )
        self.assertEqual(len(local_suppression_comments), 1)
        local_suppression_comment = local_suppression_comments[0]

        # First code unneeded.
        unneeded_codes = ["UsedRule"]
        new_comment_lines = _compose_new_comment(
            local_suppression_comment, unneeded_codes, 1
        )
        expected_new_lines = [
            "# lint-fixme: UsedRule2: reason... reason",
            "# lint: continued",
        ]
        self.assertEqual(new_comment_lines, expected_new_lines)

        # Second code unneeded.
        unneeded_codes = ["UsedRule2"]
        new_comment_lines = _compose_new_comment(
            local_suppression_comment, unneeded_codes, 1
        )
        expected_new_lines = [
            "# lint-fixme: UsedRule: reason... reason",
            "# lint: continued",
        ]
        self.assertEqual(new_comment_lines, expected_new_lines)

        # Both codes unneded.
        unneeded_codes = ["UsedRule", "UsedRule2"]
        new_comment_lines = _compose_new_comment(
            local_suppression_comment, unneeded_codes, 1
        )
        expected_new_lines = []
        self.assertEqual(new_comment_lines, expected_new_lines)

        # Both codes needed.
        unneeded_codes = []
        new_comment_lines = _compose_new_comment(
            local_suppression_comment, unneeded_codes, 1
        )
        expected_new_lines = [
            "# lint-fixme: UsedRule, UsedRule2: reason...",
            "# lint: reason continued",
        ]
        self.assertEqual(new_comment_lines, expected_new_lines)



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from libcst.testing.utils import UnitTest

from fixit.rules.import_constraints import _ImportConfig


class ImportConstraintsRuleConfigTest(UnitTest):
    def test_at_least_one_rule(self) -> None:
        # Settings must have at least one rule
        with self.assertRaisesRegex(ValueError, "at least one"):
            _ImportConfig.from_config({})

        with self.assertRaisesRegex(ValueError, "at least one"):
            _ImportConfig.from_config({"rules": []})

    def test_wildcard_rule_last(self) -> None:
        with self.assertRaisesRegex(
            ValueError, "The last rule must be a wildcard rule"
        ):
            _ImportConfig.from_config({"rules": [["module_name", "allow"]]})

        with self.assertRaisesRegex(
            ValueError, "The last rule must be a wildcard rule"
        ):
            _ImportConfig.from_config(
                {"rules": [["*", "deny"], ["module_name", "allow"]]}
            )

        with self.assertRaisesRegex(
            ValueError, "Only the last rule can be a wildcard rule"
        ):
            _ImportConfig.from_config({"rules": [["*", "deny"], ["*", "allow"]]})

    def test_allow_or_deny(self) -> None:
        with self.assertRaisesRegex(ValueError, "not a valid RuleAction"):
            _ImportConfig.from_config(
                {"rules": [["module_name", "wut"], ["*", "deny"]]}
            )

    def test_non_boolean(self) -> None:
        with self.assertRaisesRegex(ValueError, "value must be 'True' or 'False'"):
            _ImportConfig.from_config({"ignore_tests": "blah"})
        with self.assertRaisesRegex(ValueError, "value must be 'True' or 'False'"):
            _ImportConfig.from_config({"ignore_types": "blah"})

    def test_valid(self) -> None:
        _ImportConfig.from_config(
            {
                "rules": [["a.b.c", "allow"], ["d.e.f", "allow"], ["*", "deny"]],
                "ignore_tests": True,
                "ignore_types": False,
            }
        )



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Dict, Iterator, List, Set, Tuple

import libcst as cst
import libcst.matchers as m
from libcst.metadata import QualifiedName, QualifiedNameProvider, QualifiedNameSource

from fixit import (
    CstContext,
    CstLintRule,
    InvalidTestCase as Invalid,
    ValidTestCase as Valid,
)


_ISINSTANCE = QualifiedName(
    name="builtins.isinstance", source=QualifiedNameSource.BUILTIN
)


class CollapseIsinstanceChecksRule(CstLintRule):
    """
    The built-in ``isinstance`` function, instead of a single type,
    can take a tuple of types and check whether given target suits
    any of them. Rather than chaining multiple ``isinstance`` calls
    with a boolean-or operation, a single ``isinstance`` call where
    the second argument is a tuple of all types can be used.
    """

    MESSAGE: str = (
        "Multiple isinstance calls with the same target but "
        + "different types can be collapsed into a single call "
        + "with a tuple of types."
    )

    METADATA_DEPENDENCIES = (QualifiedNameProvider,)

    VALID = [
        Valid("foo() or foo()"),
        Valid("foo(x, y) or foo(x, z)"),
        Valid("foo(x, y) or foo(x, z) or foo(x, q)"),
        Valid("isinstance() or isinstance()"),
        Valid("isinstance(x) or isinstance(x)"),
        Valid("isinstance(x, y) or isinstance(x)"),
        Valid("isinstance(x) or isinstance(x, y)"),
        Valid("isinstance(x, y) or isinstance(t, y)"),
        Valid("isinstance(x, y) and isinstance(x, z)"),
        Valid("isinstance(x, y) or isinstance(x, (z, q))"),
        Valid("isinstance(x, (y, z)) or isinstance(x, q)"),
        Valid("isinstance(x, a) or isinstance(y, b) or isinstance(z, c)"),
        Valid(
            """
            def foo():
                def isinstance(x, y):
                    return _foo_bar(x, y)
                if isinstance(x, y) or isinstance(x, z):
                    print("foo")
            """
        ),
    ]
    INVALID = [
        Invalid(
            "isinstance(x, y) or isinstance(x, z)",
            expected_replacement="isinstance(x, (y, z))",
        ),
        Invalid(
            "isinstance(x, y) or isinstance(x, z) or isinstance(x, q)",
            expected_replacement="isinstance(x, (y, z, q))",
        ),
        Invalid(
            "something or isinstance(x, y) or isinstance(x, z) or another",
            expected_replacement="something or isinstance(x, (y, z)) or another",
        ),
        Invalid(
            "isinstance(x, y) or isinstance(x, z) or isinstance(x, q) or isinstance(x, w)",
            expected_replacement="isinstance(x, (y, z, q, w))",
        ),
        Invalid(
            "isinstance(x, a) or isinstance(x, b) or isinstance(y, c) or isinstance(y, d)",
            expected_replacement="isinstance(x, (a, b)) or isinstance(y, (c, d))",
        ),
        Invalid(
            "isinstance(x, a) or isinstance(x, b) or isinstance(y, c) or isinstance(y, d) "
            + "or isinstance(z, e)",
            expected_replacement="isinstance(x, (a, b)) or isinstance(y, (c, d)) or isinstance(z, e)",
        ),
        Invalid(
            "isinstance(x, a) or isinstance(x, b) or isinstance(y, c) or isinstance(y, d) "
            + "or isinstance(z, e) or isinstance(q, f) or isinstance(q, g) or isinstance(q, h)",
            expected_replacement=(
                "isinstance(x, (a, b)) or isinstance(y, (c, d)) or isinstance(z, e)"
                + " or isinstance(q, (f, g, h))"
            ),
        ),
    ]

    def __init__(self, context: CstContext) -> None:
        super().__init__(context)
        self.seen_boolean_operations: Set[cst.BooleanOperation] = set()

    def visit_BooleanOperation(self, node: cst.BooleanOperation) -> None:
        if node in self.seen_boolean_operations:
            return None

        stack = tuple(self.unwrap(node))
        operands, targets = self.collect_targets(stack)

        # If nothing gets collapsed, just exit from this short-path
        if len(operands) == len(stack):
            return None

        replacement = None
        for operand in operands:
            if operand in targets:
                matches = targets[operand]
                if len(matches) == 1:
                    arg = cst.Arg(value=matches[0])
                else:
                    arg = cst.Arg(cst.Tuple([cst.Element(match) for match in matches]))
                operand = cst.Call(cst.Name("isinstance"), [cst.Arg(operand), arg])

            if replacement is None:
                replacement = operand
            else:
                replacement = cst.BooleanOperation(
                    left=replacement, right=operand, operator=cst.Or()
                )

        if replacement is not None:
            self.report(node, replacement=replacement)

    def unwrap(self, node: cst.BaseExpression) -> Iterator[cst.BaseExpression]:
        if m.matches(node, m.BooleanOperation(operator=m.Or())):
            bool_op = cst.ensure_type(node, cst.BooleanOperation)
            self.seen_boolean_operations.add(bool_op)
            yield from self.unwrap(bool_op.left)
            yield bool_op.right
        else:
            yield node

    def collect_targets(
        self, stack: Tuple[cst.BaseExpression, ...]
    ) -> Tuple[
        List[cst.BaseExpression], Dict[cst.BaseExpression, List[cst.BaseExpression]]
    ]:
        targets = {}
        operands = []

        for operand in stack:
            if m.matches(
                operand, m.Call(func=m.DoNotCare(), args=[m.Arg(), m.Arg(~m.Tuple())])
            ):
                call = cst.ensure_type(operand, cst.Call)
                if not QualifiedNameProvider.has_name(self, call, _ISINSTANCE):
                    operands.append(operand)
                    continue

                target, match = call.args[0].value, call.args[1].value
                for possible_target in targets:
                    if target.deep_equals(possible_target):
                        targets[possible_target].append(match)
                        break
                else:
                    operands.append(target)
                    targets[target] = [match]
            else:
                operands.append(operand)

        return operands, targets



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import unittest
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Callable, cast, Dict, Mapping, Optional, Sequence, Type, Union

from libcst.metadata import MetadataWrapper

from fixit.common.base import CstLintRule
from fixit.common.generate_pyre_fixtures import get_fixture_path
from fixit.common.report import BaseLintRuleReport
from fixit.common.utils import (
    _dedent,
    gen_type_inference_wrapper,
    InvalidTestCase,
    LintRuleCollectionT,
    ValidTestCase,
)
from fixit.rule_lint_engine import lint_file


def validate_patch(report: BaseLintRuleReport, test_case: InvalidTestCase) -> None:
    patch = report.patch
    expected_replacement = test_case.expected_replacement

    if patch is None:
        if expected_replacement is not None:
            raise AssertionError(
                "The rule for this test case has no auto-fix, but expected source was specified."
            )
        return

    if expected_replacement is None:
        raise AssertionError(
            "The rule for this test case has an auto-fix, but no expected source was specified."
        )

    expected_replacement = _dedent(expected_replacement)
    patched_code = patch.apply(_dedent(test_case.code))
    if patched_code != expected_replacement:
        raise AssertionError(
            "Auto-fix did not produce expected result.\n"
            + f"Expected:\n{expected_replacement}\n"
            + f"But found:\n{patched_code}"
        )


@dataclass(frozen=True)
class TestCasePrecursor:
    rule: Type[CstLintRule]
    test_methods: Mapping[
        str,
        Union[ValidTestCase, InvalidTestCase],
    ]
    fixture_paths: Mapping[str, Path]


class LintRuleTestCase(unittest.TestCase):
    def _test_method(
        self,
        test_case: Union[ValidTestCase, InvalidTestCase],
        rule: Type[CstLintRule],
        fixture_file: Optional[Path] = None,
    ) -> None:
        cst_wrapper: Optional[MetadataWrapper] = None
        if fixture_file is not None:
            cst_wrapper = gen_type_inference_wrapper(test_case.code, fixture_file)
        reports = lint_file(
            Path(test_case.filename),
            _dedent(test_case.code).encode("utf-8"),
            config=test_case.config,
            rules={rule},
            cst_wrapper=cst_wrapper,
        )
        if isinstance(test_case, ValidTestCase):
            self.assertEqual(
                len(reports),
                0,
                'Expected zero reports for this "valid" test case. Instead, found:\n'
                + "\n".join(str(e) for e in reports),
            )
        else:
            self.assertGreater(
                len(reports),
                0,
                'Expected a report for this "invalid" test case but `self.report` was '
                + "not called:\n"
                + test_case.code,
            )
            self.assertLessEqual(
                len(reports),
                1,
                'Expected one report from this "invalid" test case. Found multiple:\n'
                + "\n".join(str(e) for e in reports),
            )

            # pyre-fixme[16]: `Collection` has no attribute `__getitem__`.
            report = reports[0]

            if not (test_case.line is None or test_case.line == report.line):
                raise AssertionError(
                    f"Expected line: {test_case.line} but found line: {report.line}"
                )

            if not (test_case.column is None or test_case.column == report.column):
                raise AssertionError(
                    f"Expected column: {test_case.column} but found column: {report.column}"
                )
            kind = test_case.kind if test_case.kind is not None else rule.__name__
            if kind != report.code:
                raise AssertionError(
                    f"Expected:\n    {test_case.expected_str}\nBut found:\n    {report}"
                )
            if (
                test_case.expected_message is not None
                and test_case.expected_message != report.message
            ):
                raise AssertionError(
                    f"Expected message:\n    {test_case.expected_message}\nBut got:\n    {report.message}"
                )

            validate_patch(report, test_case)


def _gen_test_methods_for_rule(
    rule: Type[CstLintRule], fixture_dir: Path, rules_package: str
) -> TestCasePrecursor:
    """Aggregates all of the cases inside a single CstLintRule's VALID and INVALID attributes
    and maps them to altered names with a `test_` prefix so that 'unittest' can discover them
    later on and an index postfix so that individual tests can be selected from the command line.
    """
    valid_tcs = {}
    invalid_tcs = {}
    requires_fixtures = False
    fixture_paths: Dict[str, Path] = {}
    fixture_subdir: Path = get_fixture_path(fixture_dir, rule.__module__, rules_package)
    if issubclass(rule, CstLintRule):

        if rule.requires_metadata_caches():
            requires_fixtures = True
        if hasattr(rule, "VALID"):
            # pyre-fixme[16]: `CstLintRule` has no attribute `VALID`.
            for idx, test_case in enumerate(getattr(rule, "VALID")):
                name = f"test_VALID_{idx}"
                valid_tcs[name] = test_case
                if requires_fixtures:
                    fixture_paths[name] = (
                        fixture_subdir / f"{rule.__name__}_VALID_{idx}.json"
                    )
        if hasattr(rule, "INVALID"):
            # pyre-fixme[16]: `CstLintRule` has no attribute `INVALID`.
            for idx, test_case in enumerate(getattr(rule, "INVALID")):
                name = f"test_INVALID_{idx}"
                invalid_tcs[name] = test_case
                if requires_fixtures:
                    fixture_paths[name] = (
                        fixture_subdir / f"{rule.__name__}_INVALID_{idx}.json"
                    )
    return TestCasePrecursor(
        rule=rule,
        test_methods={**valid_tcs, **invalid_tcs},
        fixture_paths=fixture_paths,
    )


def _gen_all_test_methods(
    rules: LintRuleCollectionT, fixture_dir: Path, rules_package: str
) -> Sequence[TestCasePrecursor]:
    """
    Converts all passed-in lint rules to type `TestCasePrecursor` to ease further TestCase
    creation later on.
    """
    cases = []
    for rule in rules:
        if not issubclass(rule, CstLintRule):
            continue
        test_cases_for_rule = _gen_test_methods_for_rule(
            cast(Type[CstLintRule], rule), fixture_dir, rules_package
        )
        cases.append(test_cases_for_rule)
    return cases


def add_lint_rule_tests_to_module(
    module_attrs: Dict[str, Any],
    rules: LintRuleCollectionT,
    test_case_type: Type[unittest.TestCase] = LintRuleTestCase,
    custom_test_method_name: str = "_test_method",
    fixture_dir: Path = Path(""),
    rules_package: str = "",
) -> None:
    """
    Generates classes inheriting from `unittest.TestCase` from the data available in `rules` and adds these to module_attrs.
    The goal is to facilitate unit test discovery by Python's `unittest` framework. This will provide the capability of
    testing your lint rules by running commands such as `python -m unittest <your testing module name>`.

    module_attrs: A dictionary of attributes we want to add these test cases to. If adding to a module, you can pass `globals()` as the argument.

    rules: A collection of classes extending `CstLintRule` to be converted to test cases.

    test_case_type: A class extending Python's `unittest.TestCase` that implements a custom test method for testing lint rules to serve as a stencil for test cases.
    New classes will be generated, and named after each lint rule. They will inherit directly from the class passed into `test_case_type`.
    If argument is omitted, will default to the `LintRuleTestCase` class from fixit.common.testing.

    custom_test_method_name: A member method of the class passed into `test_case_type` parameter that contains the logic around asserting success or failure of
    CstLintRule's `ValidTestCase` and `InvalidTestCase` test cases. The method will be dynamically renamed to `test_<VALID/INVALID>_<test case index>` for discovery
    by unittest. If argument is omitted, `add_lint_rule_tests_to_module` will look for a test method named `_test_method` member of `test_case_type`.

    fixture_dir: The directory in which fixture files for the passed rules live. Necessary only if any lint rules require fixture data for testing.

    rules_package: The name of the rules package. This will be used during the search for fixture files and provides insight into the structure of the fixture directory.
    The structure of the fixture directory is automatically assumed to mirror the structure of the rules package, eg: `<rules_package>.submodule.module.rule_class` should
    have fixture files in `<fixture_dir>/submodule/module/rule_class/`.
    """
    for test_case in _gen_all_test_methods(rules, fixture_dir, rules_package):
        rule_name = test_case.rule.__name__
        # pyre-fixme[24]: Generic type `Callable` expects 2 type parameters.
        test_methods_to_add: Dict[str, Callable] = {}

        for test_method_name, test_method_data in test_case.test_methods.items():
            fixture_file = test_case.fixture_paths.get(test_method_name)

            def test_method(
                self: Type[unittest.TestCase],
                data: Union[ValidTestCase, InvalidTestCase] = test_method_data,
                rule: Type[CstLintRule] = test_case.rule,
                fixture_file: Optional[str] = fixture_file,
            ) -> None:
                return getattr(self, custom_test_method_name)(data, rule, fixture_file)

            # pyre-fixme[16]: Anonymous callable has no attribute `__name__`.
            test_method.__name__ = test_method_name
            test_methods_to_add[test_method_name] = test_method

        test_case_class = type(rule_name, (test_case_type,), test_methods_to_add)
        module_attrs[rule_name] = test_case_class



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import abc
import ast
from pathlib import Path
from pickle import PicklingError
from typing import Collection, Optional, Sequence, Union

import libcst as cst

from fixit.common.autofix import LintPatch


class BaseLintRuleReport(abc.ABC):
    """
    Represents a lint violation. This is generated by calling `self.context.report`
    in your lint rule, and is saved to the context's `reports` list.
    """

    file_path: Path
    code: str
    message: str
    # This is the line/column where the lint rule reported the violation. `arc lint` may
    # report a different line/column when a patch is applied because it requires that
    # the start of the patch is the same as the reported line/column.
    line: int
    column: int

    def __init__(
        self, *, file_path: Path, code: str, message: str, line: int, column: int
    ) -> None:
        self.file_path = file_path
        self.code = code
        self.message = message
        self.line = line
        self.column = column

    @property
    def patch(self) -> Optional[LintPatch]:
        return None

    def __repr__(self) -> str:
        return f"{self.line}:{self.column}: {self.code} {self.message}"

    # pyre-fixme[15]: `__reduce__` overrides method defined in `object` inconsistently.
    def __reduce__(self) -> None:
        raise PicklingError(
            "Lint rule reports are potentially very complex objects. They can contain "
            + "a syntax tree or an entire module's source code. They should not be "
            + "pickled (or returned by a multiprocessing worker). Instead, extract "
            + "the fields you care about, and pickle those."
        )


class AstLintRuleReport(BaseLintRuleReport):
    def __init__(
        self,
        *,
        file_path: Path,
        node: ast.AST,
        code: str,
        message: str,
        line: int,
        column: int,
    ) -> None:
        super().__init__(
            file_path=file_path, code=code, message=message, line=line, column=column
        )
        self.node = node


class CstLintRuleReport(BaseLintRuleReport):
    def __init__(
        self,
        *,
        file_path: Path,
        node: cst.CSTNode,
        code: str,
        message: str,
        line: int,
        column: int,
        module: cst.MetadataWrapper,
        module_bytes: bytes,
        replacement_node: Optional[
            Union[cst.CSTNode, cst.FlattenSentinel, cst.RemovalSentinel]
        ] = None,
    ) -> None:
        super().__init__(
            file_path=file_path, code=code, message=message, line=line, column=column
        )
        self.node = node
        self.module = module
        self.module_bytes = module_bytes
        self.replacement_node = replacement_node
        self._cached_patch: Optional[LintPatch] = None

    # Ideally this would use functools.cached_property, but that's only in py3.8+.
    @property
    def patch(self) -> Optional[LintPatch]:
        """
        Computes and returns a `LintPatch` object.
        """
        replacement_node = self.replacement_node
        if replacement_node is None:
            return None
        cached = self._cached_patch
        if cached is None:
            cached = LintPatch.get(
                wrapper=self.module,
                original_node=self.node,
                replacement_node=replacement_node,
            ).minimize()
            self._cached_patch = cached
        return cached


class LintFailureReportBase(abc.ABC):
    """An implementation needs to be a dataclass."""

    @staticmethod
    @abc.abstractmethod
    def create_reports(
        path: Path, exception_traceback: str, **kwargs: object
    ) -> Sequence["LintFailureReportBase"]:
        ...


class LintSuccessReportBase(abc.ABC):
    """An implementation needs to be a dataclass."""

    @staticmethod
    @abc.abstractmethod
    def create_reports(
        path: Path, reports: Collection[BaseLintRuleReport], **kwargs: object
    ) -> Sequence["LintSuccessReportBase"]:
        ...



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Optional

import libcst as cst
import libcst.matchers as m

from fixit import CstLintRule, InvalidTestCase as Invalid, ValidTestCase as Valid


class NoStaticIfConditionRule(CstLintRule):
    """
    Discourages ``if`` conditions which evaluate to a static value (e.g. ``or True``, ``and False``, etc).
    """

    MESSAGE: str = (
        "Your if condition appears to evaluate to a static value (e.g. `or True`, `and False`). "
        + "Please double check this logic and if it is actually temporary debug code."
    )
    VALID = [
        Valid(
            """
            if my_func() or not else_func():
                pass
            """
        ),
        Valid(
            """
            if function_call(True):
                pass
            """
        ),
        Valid(
            """
            # ew who would this???
            def true():
                return False
            if true() and else_call():  # True or False
                pass
            """
        ),
        Valid(
            """
            # ew who would this???
            if False or some_func():
                pass
            """
        ),
    ]
    INVALID = [
        Invalid(
            """
            if True:
                do_something()
            """,
        ),
        Invalid(
            """
            if crazy_expression or True:
                do_something()
            """,
        ),
        Invalid(
            """
            if crazy_expression and False:
                do_something()
            """,
        ),
        Invalid(
            """
            if crazy_expression and not True:
                do_something()
            """,
        ),
        Invalid(
            """
            if crazy_expression or not False:
                do_something()
            """,
        ),
        Invalid(
            """
            if crazy_expression or (something() or True):
                do_something()
            """,
        ),
        Invalid(
            """
            if crazy_expression and (something() and (not True)):
                do_something()
            """,
        ),
        Invalid(
            """
            if crazy_expression and (something() and (other_func() and not True)):
                do_something()
            """,
        ),
        Invalid(
            """
            if (crazy_expression and (something() and (not True))) or True:
                do_something()
            """,
        ),
        Invalid(
            """
            async def some_func() -> none:
                if (await expression()) and False:
                    pass
            """,
        ),
    ]

    @classmethod
    def _extract_static_bool(cls, node: cst.BaseExpression) -> Optional[bool]:
        if m.matches(node, m.Call()):
            # cannot reason about function calls
            return None
        if m.matches(node, m.UnaryOperation(operator=m.Not())):
            sub_value = cls._extract_static_bool(
                cst.ensure_type(node, cst.UnaryOperation).expression
            )
            if sub_value is None:
                return None
            return not sub_value

        if m.matches(node, m.Name("True")):
            return True

        if m.matches(node, m.Name("False")):
            return False

        if m.matches(node, m.BooleanOperation()):
            node = cst.ensure_type(node, cst.BooleanOperation)
            left_value = cls._extract_static_bool(node.left)
            right_value = cls._extract_static_bool(node.right)
            if m.matches(node.operator, m.Or()):
                if right_value is True or left_value is True:
                    return True

            if m.matches(node.operator, m.And()):
                if right_value is False or left_value is False:
                    return False

        return None

    def visit_If(self, node: cst.If) -> None:
        if self._extract_static_bool(node.test) in {True, False}:
            self.report(node)



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Set

import libcst as cst
import libcst.matchers as m
from libcst.metadata import QualifiedNameProvider

from fixit import (
    CstContext,
    CstLintRule,
    InvalidTestCase as Invalid,
    ValidTestCase as Valid,
)


class NoStringTypeAnnotationRule(CstLintRule):
    """
    Enforce the use of type identifier instead of using string type hints for simplicity and better syntax highlighting.
    Starting in Python 3.7, ``from __future__ import annotations`` can postpone evaluation of type annotations
    `PEP 563 <https://www.python.org/dev/peps/pep-0563/#forward-references>`_
    and thus forward references no longer need to use string annotation style.
    """

    MESSAGE = "String type hints are no longer necessary in Python, use the type identifier directly."

    VALID = [
        # Usage of a Class for instantiation and typing.
        Valid(
            """
            from a.b import Class

            def foo() -> Class:
                return Class()
            """
        ),
        Valid(
            """
            import typing
            from a.b import Class

            def foo() -> typing.Type[Class]:
                return Class
            """
        ),
        Valid(
            """
            import typing
            from a.b import Class
            from c import func

            def foo() -> typing.Optional[typing.Type[Class]]:
                return Class if func() else None
            """
        ),
        Valid(
            """
            from a.b import Class

            def foo(arg: Class) -> None:
                pass

            foo(Class())
            """
        ),
        Valid(
            """
            from a.b import Class

            module_var: Class = Class()
            """
        ),
        Valid(
            """
            from typing import Literal

            def foo() -> Literal["a", "b"]:
                return "a"
            """
        ),
        Valid(
            """
            import typing

            def foo() -> typing.Optional[typing.Literal["a", "b"]]:
                return "a"
            """
        ),
    ]

    INVALID = [
        # Using string type hints isn't needed
        Invalid(
            """
            from __future__ import annotations

            from a.b import Class

            def foo() -> "Class":
                return Class()
            """,
            line=5,
            expected_replacement="""
            from __future__ import annotations

            from a.b import Class

            def foo() -> Class:
                return Class()
            """,
        ),
        Invalid(
            """
            from __future__ import annotations

            from a.b import Class

            async def foo() -> "Class":
                return await Class()
            """,
            line=5,
            expected_replacement="""
            from __future__ import annotations

            from a.b import Class

            async def foo() -> Class:
                return await Class()
            """,
        ),
        Invalid(
            """
            from __future__ import annotations

            import typing
            from a.b import Class

            def foo() -> typing.Type["Class"]:
                return Class
            """,
            line=6,
            expected_replacement="""
            from __future__ import annotations

            import typing
            from a.b import Class

            def foo() -> typing.Type[Class]:
                return Class
            """,
        ),
        Invalid(
            """
            from __future__ import annotations

            import typing
            from a.b import Class
            from c import func

            def foo() -> Optional[typing.Type["Class"]]:
                return Class if func() else None
            """,
            line=7,
            expected_replacement="""
            from __future__ import annotations

            import typing
            from a.b import Class
            from c import func

            def foo() -> Optional[typing.Type[Class]]:
                return Class if func() else None
            """,
        ),
        Invalid(
            """
            from __future__ import annotations

            from a.b import Class

            def foo(arg: "Class") -> None:
                pass

            foo(Class())
            """,
            line=5,
            expected_replacement="""
            from __future__ import annotations

            from a.b import Class

            def foo(arg: Class) -> None:
                pass

            foo(Class())
            """,
        ),
        Invalid(
            """
            from __future__ import annotations

            from a.b import Class

            module_var: "Class" = Class()
            """,
            line=5,
            expected_replacement="""
            from __future__ import annotations

            from a.b import Class

            module_var: Class = Class()
            """,
        ),
        Invalid(
            """
            from __future__ import annotations

            import typing
            from typing_extensions import Literal
            from a.b import Class

            def foo() -> typing.Tuple[Literal["a", "b"], "Class"]:
                return Class()
            """,
            line=7,
            expected_replacement="""
            from __future__ import annotations

            import typing
            from typing_extensions import Literal
            from a.b import Class

            def foo() -> typing.Tuple[Literal["a", "b"], Class]:
                return Class()
            """,
        ),
    ]

    def __init__(self, context: CstContext) -> None:
        super().__init__(context)
        self.in_annotation: Set[cst.Annotation] = set()
        self.in_literal: Set[cst.Subscript] = set()
        self.has_future_annotations_import = False

    def visit_ImportFrom(self, node: cst.ImportFrom) -> None:
        if m.matches(
            node,
            m.ImportFrom(
                module=m.Name("__future__"),
                names=[
                    m.ZeroOrMore(),
                    m.ImportAlias(name=m.Name("annotations")),
                    m.ZeroOrMore(),
                ],
            ),
        ):
            self.has_future_annotations_import = True

    def visit_Annotation(self, node: cst.Annotation) -> None:
        self.in_annotation.add(node)

    def leave_Annotation(self, original_node: cst.Annotation) -> None:
        self.in_annotation.remove(original_node)

    def visit_Subscript(self, node: cst.Subscript) -> None:
        if not self.has_future_annotations_import:
            return
        if self.in_annotation:
            if m.matches(
                node,
                m.Subscript(
                    metadata=m.MatchMetadataIfTrue(
                        QualifiedNameProvider,
                        lambda qualnames: any(
                            n.name == "typing_extensions.Literal" for n in qualnames
                        ),
                    )
                ),
                metadata_resolver=self.context.wrapper,
            ):
                self.in_literal.add(node)

    def leave_Subscript(self, original_node: cst.Subscript) -> None:
        if not self.has_future_annotations_import:
            return
        if original_node in self.in_literal:
            self.in_literal.remove(original_node)

    def visit_SimpleString(self, node: cst.SimpleString) -> None:
        if not self.has_future_annotations_import:
            return
        if self.in_annotation and not self.in_literal:
            # This is not allowed past Python3.7 since it's no longer necessary.
            self.report(
                node,
                replacement=cst.parse_expression(
                    node.evaluated_value,
                    config=self.context.wrapper.module.config_for_parsing,
                ),
            )



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import abc
import ast
import io
import tokenize
from pathlib import Path
from typing import Iterable, Optional

from fixit.common.report import BaseLintRuleReport


class PseudoContext:
    """
    Contains information about the file that `PseudoLintRule.lint_file` should evaluate.
    """

    def __init__(
        self,
        file_path: Path,
        source: bytes,
        tokens: Optional[Iterable[tokenize.TokenInfo]] = None,
        ast_tree: Optional[ast.Module] = None,
    ) -> None:
        self.file_path: Path = file_path
        self.source: bytes = source
        self._tokens: Optional[Iterable[tokenize.TokenInfo]] = tokens
        self._ast_tree: Optional[ast.Module] = ast_tree

    @property
    def tokens(self) -> Iterable[tokenize.TokenInfo]:
        tokens = self._tokens
        if tokens is not None:
            return tokens
        tokens = tuple(tokenize.tokenize(io.BytesIO(self.source).readline))
        self._tokens = tokens
        return tokens

    @property
    def ast_tree(self) -> ast.Module:
        ast_tree = self._ast_tree
        if ast_tree is not None:
            return ast_tree
        ast_tree = ast.parse(self.source)
        self._ast_tree = ast_tree
        return ast_tree


class PseudoLintRule(abc.ABC):
    """
    Represents a lint rule (or a group of lint rules) that can't be represented by a
    normal lint rule. These "pseudo" lint rules receive information about the file from
    the `PsuedoContext`.

    This API is much more flexible than the normal lint rule API, but that comes at a
    (potentially large) performance cost. Because the lint framework does not control
    traversal of the syntax tree, it cannot batch the execution of these rules alongside
    other lint rules.

    This API is used for compatibility with Flake8 rules.
    """

    def __init__(self, context: PseudoContext) -> None:
        self.context: PseudoContext = context

    @abc.abstractmethod
    def lint_file(self) -> Iterable[BaseLintRuleReport]:
        ...



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from pathlib import Path

import libcst as cst
from libcst.testing.utils import data_provider, UnitTest

from fixit import rule_lint_engine
from fixit.common.base import CstLintRule, LintConfig


class BadCallCstLintRule(CstLintRule):
    def visit_Call(self, node: cst.Call) -> None:
        func = node.func
        if isinstance(func, cst.Name) and func.value == "bad_call":
            self.report(node, "You made a bad call!")


class ParenthesizeAttributeLintRule(CstLintRule):
    """
    Transforms the following code:

        obj.attr.another_attr
    into:
        ((obj.attr).another_attr)
    This serves as an easy test case for overlapping lint fixes.
    """

    def visit_Attribute(self, node: cst.Attribute) -> None:
        rule_config = self.context.config.rule_config
        parenthesize_attribute_config = rule_config.get(self.__class__.__name__, {})
        if isinstance(
            parenthesize_attribute_config, dict
        ) and parenthesize_attribute_config.get("disabled", False):
            return
        if len(node.lpar) == 0:
            new_node = node.with_changes(
                lpar=[cst.LeftParen()], rpar=[cst.RightParen()]
            )
            self.report(
                node,
                "All attributes should be parenthesized.",
                replacement=new_node,
            )


class RuleLintEngineTest(UnitTest):
    @data_provider(
        {
            "good_call": {
                "source": b"good_call()\n",
                "use_ignore_byte_markers": False,
                "use_ignore_comments": False,
                "expected_report_count": 0,
                "use_noqa": False,
            },
            "bad_call": {
                "source": b"bad_call()\n",
                "use_ignore_byte_markers": False,
                "use_ignore_comments": False,
                "expected_report_count": 1,
                "use_noqa": False,
            },
            "multiple_bad_calls": {
                "source": b"bad_call()\nbad_call()\n",
                "use_ignore_byte_markers": False,
                "use_ignore_comments": False,
                "expected_report_count": 2,
                "use_noqa": False,
            },
            "bad_call_generated": {
                "source": b"'''@gen" + b"erated'''\nbad_call()",
                "use_ignore_byte_markers": True,
                "use_ignore_comments": False,
                "expected_report_count": 0,
                "use_noqa": False,
            },
            "bad_call_noqa": {
                "source": b"bad_call()  # noqa\n",
                "use_ignore_byte_markers": False,
                "use_ignore_comments": True,
                "expected_report_count": 0,
                "use_noqa": True,
            },
            "bad_call_noqa_mixed": {
                "source": b"bad_call()  # noqa\nbad_call()  # missing noqa comment\n",
                "use_ignore_byte_markers": False,
                "use_ignore_comments": True,
                "expected_report_count": 1,
                "use_noqa": True,
            },
            "bad_call_noqa_file": {
                "source": b"# noqa-file: BadCallCstLintRule: Test case\nbad_call()\nbad_call()\n",
                "use_ignore_byte_markers": False,
                "use_ignore_comments": True,
                "expected_report_count": 0,
                "use_noqa": True,
            },
        }
    )
    def test_lint_file(
        self,
        *,
        source: bytes,
        use_ignore_byte_markers: bool,
        use_ignore_comments: bool,
        expected_report_count: int,
        use_noqa: bool,
    ) -> None:
        reports = rule_lint_engine.lint_file(
            Path("dummy_filename.py"),
            source,
            use_ignore_byte_markers=use_ignore_byte_markers,
            use_ignore_comments=use_ignore_comments,
            config=LintConfig(use_noqa=use_noqa),
            rules={BadCallCstLintRule},
        )
        self.assertEqual(len(reports), expected_report_count)

    def test_lint_file_with_config(self) -> None:
        source = b"obj.attr.another_attr\n"
        config = LintConfig(
            rule_config={"ParenthesizeAttributeLintRule": {"disabled": True}},
        )

        reports = rule_lint_engine.lint_file(
            Path("dummy_file.py"),
            source,
            config=config,
            rules={ParenthesizeAttributeLintRule},
        )
        # Expect no reports cause disabled set to True
        self.assertEqual(len(reports), 0)

        config = LintConfig(
            rule_config={"ParenthesizeAttributeLintRule": {"disabled": False}}
        )
        reports = rule_lint_engine.lint_file(
            Path("dummy_file.py"),
            source,
            config=config,
            rules={ParenthesizeAttributeLintRule},
        )
        self.assertEqual(len(reports), 2)

    def test_lint_file_and_apply_patches(self) -> None:
        source = b"obj.attr.another_attr\n"
        expected_output = b"((obj.attr).another_attr)\n"

        result = rule_lint_engine.lint_file_and_apply_patches(
            Path("dummy_filename.py"),
            source,
            config=LintConfig(),
            rules={ParenthesizeAttributeLintRule},
        )
        self.assertEqual(len(result.reports), 2)
        self.assertEqual(result.patched_source, expected_output)



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Set

import libcst
from libcst.metadata import QualifiedNameProvider, ScopeProvider

from fixit import (
    CstContext,
    CstLintRule,
    InvalidTestCase as Invalid,
    ValidTestCase as Valid,
)


REPLACE_BUILTIN_TYPE_ANNOTATION: str = (
    "You are using builtins.{builtin_type} as a type annotation "
    + "but the type system doesn't recognize it as a valid type."
    + " You should use typing.{correct_type} instead."
)

BUILTINS_TO_REPLACE: Set[str] = {"dict", "list", "set", "tuple"}
QUALIFIED_BUILTINS_TO_REPLACE: Set[str] = {f"builtins.{s}" for s in BUILTINS_TO_REPLACE}


class UseTypesFromTypingRule(CstLintRule):
    """
    Enforces the use of types from the ``typing`` module in type annotations in place of ``builtins.{builtin_type}``
    since the type system doesn't recognize the latter as a valid type.
    """

    METADATA_DEPENDENCIES = (
        QualifiedNameProvider,
        ScopeProvider,
    )
    VALID = [
        Valid(
            """
            def fuction(list: List[str]) -> None:
                pass
            """
        ),
        Valid(
            """
            def function() -> None:
                thing: Dict[str, str] = {}
            """
        ),
        Valid(
            """
            def function() -> None:
                thing: Tuple[str]
            """
        ),
        Valid(
            """
            from typing import Dict, List
            def function() -> bool:
                    return Dict == List
            """
        ),
        Valid(
            """
            from typing import List as list
            from graphene import List

            def function(a: list[int]) -> List[int]:
                    return []
            """
        ),
    ]
    INVALID = [
        Invalid(
            """
            from typing import List
            def whatever(list: list[str]) -> None:
                pass
            """,
            expected_replacement="""
            from typing import List
            def whatever(list: List[str]) -> None:
                pass
            """,
        ),
        Invalid(
            """
            def function(list: list[str]) -> None:
                pass
            """,
        ),
        Invalid(
            """
            def func() -> None:
                thing: dict[str, str] = {}
            """,
        ),
        Invalid(
            """
            def func() -> None:
                thing: tuple[str]
            """,
        ),
        Invalid(
            """
            from typing import Dict
            def func() -> None:
                thing: dict[str, str] = {}
            """,
            expected_replacement="""
            from typing import Dict
            def func() -> None:
                thing: Dict[str, str] = {}
            """,
        ),
    ]

    def __init__(self, context: CstContext) -> None:
        super().__init__(context)
        self.annotation_counter: int = 0

    def visit_Annotation(self, node: libcst.Annotation) -> None:
        self.annotation_counter += 1

    def leave_Annotation(self, original_node: libcst.Annotation) -> None:
        self.annotation_counter -= 1

    def visit_Name(self, node: libcst.Name) -> None:
        # Avoid a false-positive in this scenario:
        #
        # ```
        # from typing import List as list
        # from graphene import List
        # ```
        qualified_names = self.get_metadata(QualifiedNameProvider, node, set())

        is_builtin_type = node.value in BUILTINS_TO_REPLACE and all(
            qualified_name.name in QUALIFIED_BUILTINS_TO_REPLACE
            for qualified_name in qualified_names
        )

        if self.annotation_counter > 0 and is_builtin_type:
            correct_type = node.value.title()
            scope = self.get_metadata(ScopeProvider, node)
            replacement = None
            if scope is not None and correct_type in scope:
                replacement = node.with_changes(value=correct_type)
            self.report(
                node,
                REPLACE_BUILTIN_TYPE_ANNOTATION.format(
                    builtin_type=node.value, correct_type=correct_type
                ),
                replacement=replacement,
            )



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import textwrap
from typing import ClassVar

from fixit.common.report import BaseLintRuleReport


class LintRuleReportFormatter:
    """
    A pretty-printer for BaseLintRuleReport objects.

    The methods in this class can be overridden to tweak the formatting.
    """

    # If your terminal window is narrower than this, we'll give up and just wrap the
    # content to this width instead.
    #
    # This is necessary since some elements (e.g. an indent) have a fixed width.
    MIN_WIDTH: ClassVar[int] = 5
    width: int
    compact: bool

    def __init__(self, width: int, compact: bool = False) -> None:
        self.width = max(width, self.MIN_WIDTH)
        self.compact = compact

    def _format_position(self, report: BaseLintRuleReport) -> str:
        return f"{report.file_path}:{report.line}:{report.column}"

    def _format_header(self, report: BaseLintRuleReport) -> str:
        return self._format_position(report)

    def _format_details_raw(self, report: BaseLintRuleReport) -> str:
        return f"{report.code}: {report.message}"

    def _format_details(self, report: BaseLintRuleReport) -> str:
        lines = self._format_details_raw(report).split("\n")
        wrapped_lines = []
        for line in lines:
            if line == "":
                wrapped_lines.append(line)
            else:
                wrapped_lines.extend(
                    textwrap.wrap(
                        line,
                        self.width,
                        initial_indent=(" " * 4),
                        subsequent_indent=(" " * 4),
                    )
                )
        return "\n".join(wrapped_lines)

    def format(self, report: BaseLintRuleReport) -> str:
        if self.compact:
            return f"{self._format_header(report)}"
        else:
            return f"{self._format_header(report)}\n{self._format_details(report)}"


def format_warning(message: str, width: int, prefix: str = "Warning: ") -> str:
    return "\n".join(
        textwrap.wrap(
            message,
            max(width, len(prefix) + 1),
            initial_indent=prefix,
            subsequent_indent=(" " * len(prefix)),
        )
    )



# Copyright (c) Meta Platforms, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import logging
from pathlib import Path
from typing import Generator, Iterable, List

import trailrunner

from .config import generate_config
from .engine import collect_rules, collect_violations
from .types import Config, FileContent, LintViolation, Result

logger = logging.getLogger(__name__)


def _make_result(path: Path, violations: Iterable[LintViolation]) -> Iterable[Result]:
    try:
        for violation in violations:
            yield Result(path, violation)
    except Exception as error:
        # TODO: this is not the right place to catch errors
        logger.debug("Exception while linting", exc_info=error)
        yield Result(path, violation=None, error=error)


def fixit_bytes(
    path: Path,
    content: FileContent,
    *,
    config: Config,
) -> Generator[Result, None, None]:
    """
    Lint raw bytes content representing a single path, using the given configuration.
    """
    rules = collect_rules(config)
    yield from _make_result(path, collect_violations(content, rules))


def fixit_file(
    path: Path,
) -> Generator[Result, None, None]:
    """
    Lint a single file on disk.
    """
    path = path.resolve()

    try:
        content: FileContent = path.read_bytes()
        config = generate_config(path)

        yield from fixit_bytes(path, content, config=config)

    except Exception as error:
        logger.debug("Exception while fixit_file", exc_info=error)
        yield Result(path, violation=None, error=error)


def fixit_paths(
    paths: Iterable[Path],
) -> Generator[Result, None, None]:
    """
    Lint multiple files or directories, recursively expanding each path.
    """
    if not paths:
        return

    expanded_paths: List[Path] = []
    for path in paths:
        expanded_paths.extend(trailrunner.walk(path))

    for path in expanded_paths:
        yield from fixit_file(path)



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# Usage:
#
#   $ python -m fixit.cli.add_new_rule --help
#   $ python -m fixit.cli.add_new_rule
#   $ python -m fixit.cli.add_new_rule --path fixit/rules/new_rule.py --name rule_name

import argparse
import sys
from pathlib import Path
from typing import Optional

from libcst.codemod._cli import invoke_formatter

from fixit.cli.utils import snake_to_camelcase
from fixit.common.config import get_lint_config


_LICENCE = """# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""

_IMPORTS = """
import libcst as cst
import libcst.matchers as m

from fixit import CstLintRule, InvalidTestCase as Invalid, ValidTestCase as Valid

"""

_TO_DOS = """
\"""
This is a model rule file for adding a new rule to fixit module
\"""

"""

_RULE_CLASS = """
class {class_name}Rule(CstLintRule):
    \"""
    docstring or new_rule description
    \"""

    MESSAGE = "Enter rule description message"

    VALID = [Valid("'example'")]

    INVALID = [Invalid("'example'")]
"""


def is_path_exists(path: str) -> Path:
    """Check for valid path, if yes, return `Path` else raise `Error`"""
    filepath = Path(path)
    if filepath.exists():
        raise FileExistsError(f"{filepath} already exists")
    elif not filepath.parent.exists():
        raise TypeError(f"{filepath} is not a valid path, provide path with file name")
    else:
        return filepath


def is_valid_name(name: str) -> str:
    """Check for valid rule name"""
    if name.casefold().endswith(("rule", "rules")):
        raise ValueError("Please enter rule name without the suffix, 'rule' or 'Rule'")
    return snake_to_camelcase(name)


def create_rule_file(file_path: Path, rule_name: str) -> None:
    """Create a new rule file."""
    rule_name = is_valid_name(rule_name)
    context = _LICENCE + _IMPORTS + _TO_DOS + _RULE_CLASS
    updated_context = invoke_formatter(
        get_lint_config().formatter, context.format(class_name=rule_name)
    )

    with open(file_path, "w") as f:
        f.write(updated_context)

    print(f"Successfully created {file_path.name} rule file at {file_path.parent}")


def _add_arguments(parser: argparse.ArgumentParser) -> None:
    """All required arguments for `add_new_rule`"""
    parser.add_argument(
        "--path",
        type=is_path_exists,
        default=Path("fixit/rules/new.py"),
        help="Path to add rule file, defaults to fixit/rules/new.py",
    )
    parser.add_argument(
        "--name",
        type=str,
        default="",
        help="Name of the rule, defaults to `New`",
    )


def register_subparser(parser: Optional[argparse._SubParsersAction] = None) -> None:
    """Add parser or subparser for `add_new_rule` command."""
    if parser is None:
        add_rule_parser = argparse.ArgumentParser(
            description="Creates a skeleton of adding new rule file",
        )
        _add_arguments(add_rule_parser)
        sys.exit(_main(add_rule_parser.parse_args()))

    else:
        add_rule_parser = parser.add_parser(
            "add_new_rule",
            description="Creates a skeleton of adding new rule file.",
            help="Creates a skeleton of adding new rule file.",
        )
        _add_arguments(add_rule_parser)
        add_rule_parser.set_defaults(subparser_fn=_main)


def _main(args: argparse.Namespace) -> int:
    file_path = args.path
    rule_name = args.name if args.name else file_path.stem
    create_rule_file(file_path, rule_name)

    return 0


if __name__ == "__main__":
    register_subparser()



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import subprocess
from pathlib import Path
from typing import Collection, Dict, Mapping, Optional, Set, Union
from unittest.mock import MagicMock, patch

import libcst as cst
from libcst.metadata import MetadataWrapper, TypeInferenceProvider
from libcst.metadata.base_provider import ProviderT
from libcst.testing.utils import UnitTest

from fixit.cli import LintWorkers, map_paths
from fixit.common.base import CstLintRule, LintConfig, LintRuleT
from fixit.common.full_repo_metadata import FullRepoMetadataConfig, get_repo_caches
from fixit.common.report import BaseLintRuleReport
from fixit.rule_lint_engine import lint_file


SOURCE_CODE = b"class Foo: pass"


class DummyTypeDependentRule(CstLintRule):
    METADATA_DEPENDENCIES = (TypeInferenceProvider,)

    def visit_Module(self, node: cst.Module) -> None:
        self.report(node, "Dummy message")


def map_paths_operation(
    path: Path,
    rules: Set[LintRuleT],
    type_cache: Optional[Mapping[ProviderT, object]],
) -> Union[str, Collection[BaseLintRuleReport]]:
    # A top-level function to be accessible by `map_paths` from `fixit.cli`.
    cst_wrapper = None
    try:
        if type_cache is not None:
            cst_wrapper = MetadataWrapper(
                cst.parse_module(SOURCE_CODE),
                True,
                type_cache,
            )
        return lint_file(
            file_path=path,
            source=SOURCE_CODE,
            rules=rules,
            cst_wrapper=cst_wrapper,
            config=LintConfig(),
        )
    except Exception as e:
        return str(e)


class TypeInferenceTest(UnitTest):
    DUMMY_PATH: Path = Path("fake/path.py")
    DUMMY_PATH_2: Path = Path("fake/path_2.py")

    def setUp(self) -> None:
        self.mock_operation: MagicMock = MagicMock(side_effect=map_paths_operation)
        self.fake_pyre_data: object = {
            "types": [
                {
                    "location": {
                        "start": {"line": 0, "column": 0},
                        "stop": {"line": 0, "column": 0},
                    },
                    "annotation": "typing.Any",
                },
            ],
        }

    @patch("libcst.metadata.TypeInferenceProvider.gen_cache")
    def test_basic_type_inference(self, gen_cache: MagicMock) -> None:
        # We want to intercept any calls that will require the pyre engine to be running.
        gen_cache.return_value = {str(self.DUMMY_PATH): self.fake_pyre_data}
        paths = (str(path) for path in [self.DUMMY_PATH])
        type_caches: Mapping[str, Dict[ProviderT, object]] = get_repo_caches(
            paths, FullRepoMetadataConfig({TypeInferenceProvider}, 1)
        )
        paths = (path for path in type_caches.keys())

        reports = next(
            map_paths(
                operation=self.mock_operation,
                paths=(path for path in type_caches.keys()),
                config={DummyTypeDependentRule},
                workers=LintWorkers.USE_CURRENT_THREAD,
                metadata_caches=type_caches,
            )
        )

        self.mock_operation.assert_called_with(
            self.DUMMY_PATH, {DummyTypeDependentRule}, type_caches[str(self.DUMMY_PATH)]
        )

        self.assertEqual(len(reports), 1)
        self.assertEqual(reports[0].file_path, self.DUMMY_PATH)

    @patch("libcst.metadata.TypeInferenceProvider.gen_cache")
    def test_basic_type_inference_multiple_paths(self, gen_cache: MagicMock) -> None:
        paths = (str(self.DUMMY_PATH), str(self.DUMMY_PATH_2))
        gen_cache.return_value = {path: self.fake_pyre_data for path in paths}

        type_caches: Mapping[str, Dict[ProviderT, object]] = get_repo_caches(
            paths, FullRepoMetadataConfig({TypeInferenceProvider}, 1)
        )

        all_reports = map_paths(
            operation=self.mock_operation,
            paths=paths,
            config={DummyTypeDependentRule},
            workers=LintWorkers.USE_CURRENT_THREAD,
            metadata_caches=type_caches,
        )

        # Reports should be returned in order since we specified `LintWorkers.USE_CURRENT_THREAD`.
        for i, reports in enumerate(all_reports):
            self.assertEqual(len(reports), 1)
            self.assertEqual(reports[0].file_path, Path(paths[i]))

    @patch("libcst.metadata.TypeInferenceProvider.gen_cache")
    def test_type_inference_with_timeout(self, gen_cache: MagicMock) -> None:
        timeout = 1
        timeout_error = subprocess.TimeoutExpired(cmd="pyre query ...", timeout=timeout)
        gen_cache.side_effect = timeout_error

        paths = (str(self.DUMMY_PATH),)
        type_caches = get_repo_caches(
            paths, FullRepoMetadataConfig({TypeInferenceProvider}, timeout)
        )

        reports = next(
            map_paths(
                operation=self.mock_operation,
                paths=paths,
                config={DummyTypeDependentRule},
                workers=LintWorkers.USE_CURRENT_THREAD,
                metadata_caches=type_caches,
            )
        )

        # Expecting a placeholder cache to have been plugged in instead.
        self.mock_operation.assert_called_with(
            self.DUMMY_PATH,
            {DummyTypeDependentRule},
            {TypeInferenceProvider: {"types": []}},
        )

        self.assertEqual(len(reports), 1)



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import List, Union

import libcst as cst
import libcst.matchers as m

from fixit import CstLintRule, InvalidTestCase as Invalid, ValidTestCase as Valid


LineType = Union[cst.BaseSmallStatement, cst.BaseStatement]


class SortedAttributesRule(CstLintRule):
    """
    Ever wanted to sort a bunch of class attributes alphabetically?
    Well now it's easy! Just add "@sorted-attributes" in the doc string of
    a class definition and lint will automatically sort all attributes alphabetically.

    Feel free to add other methods and such -- it should only affect class attributes.
    """

    INVALID = [
        Invalid(
            """
            class MyUnsortedConstants:
                \"\"\"
                @sorted-attributes
                \"\"\"
                z = "hehehe"
                B = 'aaa234'
                A = 'zzz123'
                cab = "foo bar"
                Daaa = "banana"

                @classmethod
                def get_foo(cls) -> str:
                    return "some random thing"
           """,
            expected_replacement="""
            class MyUnsortedConstants:
                \"\"\"
                @sorted-attributes
                \"\"\"
                A = 'zzz123'
                B = 'aaa234'
                Daaa = "banana"
                cab = "foo bar"
                z = "hehehe"

                @classmethod
                def get_foo(cls) -> str:
                    return "some random thing"
           """,
        )
    ]
    VALID = [
        Valid(
            """
            class MyConstants:
                \"\"\"
                @sorted-attributes
                \"\"\"
                A = 'zzz123'
                B = 'aaa234'

            class MyUnsortedConstants:
                B = 'aaa234'
                A = 'zzz123'
           """
        )
    ]
    MESSAGE: str = "It appears you are using the @sorted-attributes directive and the class variables are unsorted. See the lint autofix suggestion."

    def visit_ClassDef(self, node: cst.ClassDef) -> None:
        doc_string = node.get_docstring()
        if not doc_string or "@sorted-attributes" not in doc_string:
            return

        found_any_assign: bool = False
        pre_assign_lines: List[LineType] = []
        assign_lines: List[LineType] = []
        post_assign_lines: List[LineType] = []

        def _add_unmatched_line(line: LineType) -> None:
            post_assign_lines.append(
                line
            ) if found_any_assign else pre_assign_lines.append(line)

        for line in node.body.body:
            if m.matches(
                line, m.SimpleStatementLine(body=[m.Assign(targets=[m.AssignTarget()])])
            ):
                found_any_assign = True
                assign_lines.append(line)
            else:
                _add_unmatched_line(line)
                continue

        sorted_assign_lines = sorted(
            assign_lines, key=lambda line: line.body[0].targets[0].target.value
        )
        if sorted_assign_lines == assign_lines:
            return
        self.report(
            node,
            replacement=node.with_changes(
                body=node.body.with_changes(
                    body=pre_assign_lines + sorted_assign_lines + post_assign_lines
                )
            ),
        )



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import io
import tokenize
from dataclasses import dataclass
from pathlib import Path
from typing import cast, Collection, List, Optional, Sequence, Type

import libcst as cst
from libcst.metadata import MetadataWrapper

from fixit.common.base import CstContext, CstLintRule, LintConfig
from fixit.common.comments import CommentInfo
from fixit.common.config import get_lint_config
from fixit.common.ignores import IgnoreInfo
from fixit.common.line_mapping import LineMappingInfo
from fixit.common.pseudo_rule import PseudoContext, PseudoLintRule
from fixit.common.report import BaseLintRuleReport
from fixit.common.unused_suppressions import RemoveUnusedSuppressionsRule
from fixit.common.utils import LintRuleCollectionT


def _detect_encoding(source: bytes) -> str:
    return tokenize.detect_encoding(io.BytesIO(source).readline)[0]


def _get_tokens(source: bytes) -> Sequence[tokenize.TokenInfo]:
    return tuple(tokenize.tokenize(io.BytesIO(source).readline))


def _visit_cst_rules_with_context(
    wrapper: MetadataWrapper, rules: Collection[Type[CstLintRule]], context: CstContext
) -> None:
    # construct the rules with the context
    rule_instances = [r(context) for r in rules]
    rule_instances = [r for r in rule_instances if not r.should_skip_file()]

    # before_visit/after_leave are used to update the context
    def before_visit(node: cst.CSTNode) -> None:
        context.node_stack.append(node)

    def after_leave(node: cst.CSTNode) -> None:
        context.node_stack.pop()

    # visit the tree, dispatching to each rule as needed
    wrapper.visit_batched(
        rule_instances, before_visit=before_visit, after_leave=after_leave
    )


def lint_file(
    file_path: Path,
    source: bytes,
    *,
    use_ignore_byte_markers: bool = True,
    use_ignore_comments: bool = True,
    config: Optional[LintConfig] = None,
    rules: LintRuleCollectionT,
    cst_wrapper: Optional[MetadataWrapper] = None,
    find_unused_suppressions: bool = False,
) -> Collection[BaseLintRuleReport]:
    """
    May raise a SyntaxError, which should be handled by the
    caller.
    """
    # Get settings from the nearest `.fixit.config.yaml` file if necessary.
    # pyre-fixme[35]: Target cannot be annotated.
    config: LintConfig = config if config is not None else get_lint_config()

    if use_ignore_byte_markers and any(
        pattern.encode() in source for pattern in config.block_list_patterns
    ):
        return []

    tokens = None
    if use_ignore_comments:
        # Don't compute tokens unless we have to, it slows down
        # `fixit.cli.run_rules`.
        #
        # `tokenize` is actually much more expensive than generating the whole AST,
        # since AST parsing is heavily optimized C, and tokenize is pure python.
        tokens = _get_tokens(source)
        ignore_info = IgnoreInfo.compute(
            comment_info=CommentInfo.compute(tokens=tokens),
            line_mapping_info=LineMappingInfo.compute(tokens=tokens),
            use_noqa=config.use_noqa,
        )
    else:
        ignore_info = None

    # Don't waste time evaluating rules that are globally ignored.
    evaluated_rules = [
        r for r in rules if ignore_info is None or ignore_info.should_evaluate_rule(r)
    ]
    # Categorize lint rules.
    cst_rules: List[Type[CstLintRule]] = []
    pseudo_rules: List[Type[PseudoLintRule]] = []
    for r in evaluated_rules:
        if issubclass(r, CstLintRule):
            cst_rules.append(cast(Type[CstLintRule], r))
        elif issubclass(r, PseudoLintRule):
            pseudo_rules.append(cast(Type[PseudoLintRule], r))

    # `self.context.report()` accumulates reports into the context object, we'll copy
    # those into our local `reports` list.
    ast_tree = None
    reports = []

    if cst_wrapper is None:
        cst_wrapper = MetadataWrapper(cst.parse_module(source), unsafe_skip_copy=True)
    if cst_rules:
        cst_context = CstContext(cst_wrapper, source, file_path, config)
        _visit_cst_rules_with_context(cst_wrapper, cst_rules, cst_context)
        reports.extend(cst_context.reports)

    if pseudo_rules:
        psuedo_context = PseudoContext(file_path, source, tokens, ast_tree)
        for pr_cls in pseudo_rules:
            # pyre-fixme[45]: Cannot instantiate abstract class `PseudoLintRule`.
            reports.extend(pr_cls(psuedo_context).lint_file())

    if ignore_info is not None:
        # filter the accumulated errors that should be suppressed and report unused suppressions
        reports = [r for r in reports if not ignore_info.should_ignore_report(r)]
        if find_unused_suppressions and cst_rules:
            # We had to make sure to call ignore_info.should_ignore_report before running our
            # RemoveUnusedSuppressionsRule because ignore_info needs to be up to date for it to work.
            # We can construct a new context since we want a fresh set of reports to append to reports.
            config.rule_config[RemoveUnusedSuppressionsRule.__name__] = {
                "ignore_info": ignore_info,
                "rules": cst_rules,
            }
            unused_suppressions_context = CstContext(
                cst_wrapper, source, file_path, config
            )
            _visit_cst_rules_with_context(
                cst_wrapper, [RemoveUnusedSuppressionsRule], unused_suppressions_context
            )
            reports.extend(unused_suppressions_context.reports)

    return reports


@dataclass(frozen=True)
class LintRuleReportsWithAppliedPatches:
    reports: Collection[BaseLintRuleReport]
    patched_source: str


def lint_file_and_apply_patches(
    file_path: Path,
    source: bytes,
    *,
    use_ignore_byte_markers: bool = True,
    use_ignore_comments: bool = True,
    config: Optional[LintConfig] = None,
    rules: LintRuleCollectionT,
    max_iter: int = 100,
    cst_wrapper: Optional[MetadataWrapper] = None,
    find_unused_suppressions: bool = False,
) -> LintRuleReportsWithAppliedPatches:
    """
    Runs `lint_file` in a loop, patching one auto-fixable report on each iteration.

    Applying a single fix at a time prevents the scenario where multiple autofixes
    combine in a way that results in invalid code.
    """
    # lint_file will fetch this if we don't, but it requires disk I/O, so let's fetch it
    # here to avoid hitting the disk inside our autofixer loop.
    config = config if config is not None else get_lint_config()

    reports = []
    fixed_reports = []
    # Avoid getting stuck in an infinite loop, cap the number of iterations at `max_iter`.
    for i in range(max_iter):
        reports = lint_file(
            file_path,
            source,
            use_ignore_byte_markers=use_ignore_byte_markers,
            use_ignore_comments=use_ignore_comments,
            config=config,
            rules=rules,
            cst_wrapper=cst_wrapper,
            find_unused_suppressions=find_unused_suppressions,
        )

        try:
            first_fixable_report = next(r for r in reports if r.patch is not None)
        except StopIteration:
            # No reports with autofix were found.
            break
        else:
            # We found a fixable report. Patch and re-run the linter on this file.
            patch = first_fixable_report.patch
            assert patch is not None
            # TODO: This is really inefficient because we're forced to decode/reencode
            # the source representation, just so that lint_file can decode the file
            # again.
            #
            # We probably need to rethink how we're representing the source code.
            encoding = _detect_encoding(source)
            source = patch.apply(source.decode(encoding)).encode(encoding)
            fixed_reports.append(first_fixable_report)

    # `reports` shouldn't contain any fixable reports at this point, so there should be
    # no overlap between `fixed_reports` and `reports`.
    return LintRuleReportsWithAppliedPatches(
        # pyre-fixme[60]: Concatenation not yet support for multiple variadic
        #  tuples: `*fixed_reports, *reports`.
        reports=(*fixed_reports, *reports),
        # pyre-fixme[6]: Expected `str` for 2nd param but got `bytes`.
        patched_source=source,
    )



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import distutils.spawn


try:
    import importlib.resources as pkg_resources
except ImportError:  # For <=3.6
    import importlib_resources as pkg_resources

import json
import os
import re
from dataclasses import asdict
from functools import lru_cache
from pathlib import Path
from typing import Any, Dict, Pattern, Set

import yaml
from jsonschema import validate

from fixit.common.base import LintConfig
from fixit.common.utils import import_distinct_rules_from_package, LintRuleCollectionT


LINT_CONFIG_FILE_NAME: Path = Path(".fixit.config.yaml")
LINT_CONFIG_SCHEMA_NAME: str = "config.schema.json"

# https://gitlab.com/pycqa/flake8/blob/9631dac52aa6ed8a3de9d0983c/src/flake8/defaults.py
NOQA_INLINE_REGEXP: Pattern[str] = re.compile(
    # TODO: Deprecate
    # We're looking for items that look like this:
    # ``# noqa``
    # ``# noqa: E123``
    # ``# noqa: E123,W451,F921``
    # ``# NoQA: E123,W451,F921``
    # ``# NOQA: E123,W451,F921``
    # We do not care about the ``: `` that follows ``noqa``
    # We do not care about the casing of ``noqa``
    # We want a comma-separated list of errors
    r"^# noqa(?!-file)(?:: (?P<codes>([a-zA-Z0-9]+,\s*)*[-_a-zA-Z0-9]+))?",
    re.IGNORECASE,
)
LINT_IGNORE_REGEXP: Pattern[str] = re.compile(
    # We're looking for items that look like this:
    # ``# lint-fixme: LintRuleName``
    # ``# lint-fixme: LintRuleName: Details``
    # ``# lint-fixme: LintRuleName1, LintRuleName2: Details``
    r"# lint-(ignore|fixme)"
    + r": (?P<codes>([-_a-zA-Z0-9]+,\s*)*[-_a-zA-Z0-9]+)"
    + r"(:( (?P<reason>.+)?)?)?"
)

# Skip evaluation of the given file.
# People should use `# noqa-file` or `@no- lint` instead. This is here for compatibility
# with Flake8.
FLAKE8_NOQA_FILE: Pattern[str] = re.compile(r"# flake8[:=]\s*noqa", re.IGNORECASE)

# Skip evaluation of a given rule for a given file
# `# noqa-file: LintRuleName: Some reason why we can't use this rule`
# `# noqa-file: LintRuleName,LintRuleName2: Some reason why we can't use these rules`
#
# TODO: Deprecate
NOQA_FILE_RULE: Pattern[str] = re.compile(
    r"# noqa-file: "
    + r"(?P<codes>([-_a-zA-Z0-9]+,\s*)*[-_a-zA-Z0-9]+): "
    + r"(?P<reason>.+)"
)

DEFAULT_FORMATTER = ["black", "-"]
PATH_SETTINGS = ["repo_root", "fixture_dir"]


def get_validated_settings(
    file_content: Dict[str, Any], current_dir: Path
) -> Dict[str, Any]:
    # __package__ should never be none (config.py should not be run directly)
    # But use .get() to make pyre happy
    pkg = globals().get("__package__")
    assert pkg, "No package was found, config types not validated."
    config = pkg_resources.read_text(pkg, LINT_CONFIG_SCHEMA_NAME)
    # Validates the types and presence of the keys
    schema = json.loads(config)
    validate(instance=file_content, schema=schema)

    for path_setting_name in PATH_SETTINGS:
        if path_setting_name in file_content:
            setting_value = file_content[path_setting_name]
            abspath: Path = (current_dir / setting_value).resolve()
        else:
            abspath: Path = current_dir
        # Set path setting to absolute path.
        file_content[path_setting_name] = str(abspath)

    return file_content


@lru_cache()
def get_lint_config() -> LintConfig:
    config = {}

    cwd = Path.cwd()
    for directory in (cwd, *cwd.parents):
        # Check for config file.
        # pyre-fixme[58]: `/` is not supported for operand types `object` and `Path`.
        possible_config = directory / LINT_CONFIG_FILE_NAME
        if possible_config.is_file():
            with open(possible_config, "r") as f:
                file_content = yaml.safe_load(f.read())

            if isinstance(file_content, dict):
                # pyre-fixme[6]: Expected `Path` for 2nd param but got `object`.
                config = get_validated_settings(file_content, directory)
                break

    # Find formatter executable if there is one.
    formatter_args = config.get("formatter", DEFAULT_FORMATTER)
    exe = distutils.spawn.find_executable(formatter_args[0]) or formatter_args[0]
    formatter_args[0] = os.path.abspath(exe)
    config["formatter"] = formatter_args

    # Missing settings will be populated with defaults.
    return LintConfig(**config)


def gen_config_file() -> None:
    # Generates a `.fixit.config.yaml` file with defaults in the current working dir.
    config_file = LINT_CONFIG_FILE_NAME.resolve()
    default_config_dict = asdict(LintConfig())
    with open(config_file, "w") as cf:
        yaml.dump(default_config_dict, cf)


def get_rules_from_config() -> LintRuleCollectionT:
    # Get rules from the packages specified in the lint config file, omitting block-listed rules.
    lint_config = get_lint_config()
    rules: LintRuleCollectionT = set()
    all_names: Set[str] = set()
    for package in lint_config.packages:
        rules_from_pkg = import_distinct_rules_from_package(
            package,
            lint_config.block_list_rules,
            all_names,
            lint_config.allow_list_rules,
        )
        rules.update(rules_from_pkg)
    return rules



from __future__ import annotations

from abc import ABC, abstractmethod
from collections import defaultdict

from typing import (
    Callable,
    ClassVar,
    Collection,
    Dict,
    Generic,
    Iterable,
    List,
    Optional,
    Set,
    Type,
    TypeVar,
)

from fixit.types import FileContent, LintViolation


Timings = Dict[str, int]
TimingsHook = Callable[[Timings], None]


class LintRule:
    TAGS: Set[str] = set()
    _violations: List[LintViolation] = []
    _runner: ClassVar[Type[LintRunner]]


SomeRule = TypeVar("SomeRule", bound=LintRule)


class LintRunner(ABC, Generic[SomeRule]):
    def __init__(self) -> None:
        self.timings: Timings = defaultdict(lambda: 0)

    @abstractmethod
    def collect_violations(
        self,
        source: FileContent,
        rules: Collection[SomeRule],
        timings_hook: Optional[TimingsHook] = None,
    ) -> Iterable[LintViolation]:
        pass


__all__ = ["LintRule", "LintRunner", "Timings", "TimingsHook"]



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# Usage:
#
#   $ python -m fixit.cli.insert_suppressions --help
#   $ python -m fixit.cli.insert_suppressions fixit.rules.avoid_or_in_except.AvoidOrInExceptRule
#   $ python -m fixit.cli.insert_suppressions fixit.rules.avoid_or_in_except.AvoidOrInExceptRule .

import argparse
import itertools
import shutil
import sys
import time
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import Iterable, List, Mapping, Optional, TYPE_CHECKING, Union

from libcst import parse_module, ParserSyntaxError
from libcst.codemod._cli import invoke_formatter
from libcst.metadata import MetadataWrapper

from fixit.cli import find_files, map_paths
from fixit.cli.args import (
    get_compact_parser,
    get_metadata_cache_parser,
    get_multiprocessing_parser,
    get_paths_parser,
    get_rule_parser,
    get_skip_autoformatter_parser,
)
from fixit.cli.formatter import LintRuleReportFormatter
from fixit.cli.full_repo_metadata import (
    get_metadata_caches,
    rules_require_metadata_cache,
)
from fixit.cli.utils import print_red
from fixit.common.base import LintRuleT
from fixit.common.config import get_lint_config
from fixit.common.insert_suppressions import (
    insert_suppressions,
    SuppressionComment,
    SuppressionCommentKind,
)
from fixit.common.report import BaseLintRuleReport
from fixit.rule_lint_engine import lint_file


if TYPE_CHECKING:
    from libcst.metadata.base_provider import ProviderT

DESCRIPTION: str = """Inserts `# lint-fixme` comments into a file where lint violations
are found. You should only use this tool if it's not feasible to fix the existing
violations."""

PARENTS: List[argparse.ArgumentParser] = [
    get_rule_parser(),
    get_paths_parser(),
    get_skip_autoformatter_parser(),
    get_compact_parser(),
    get_metadata_cache_parser(),
    get_multiprocessing_parser(),
]


class MessageKind(Enum):
    USE_LINT_REPORT = 1
    NO_MESSAGE = 2


@dataclass(frozen=True)
class InsertSuppressionsOpts:
    rule: LintRuleT
    skip_autoformatter: bool
    kind: SuppressionCommentKind
    message: Union[MessageKind, str]
    max_lines: int
    formatter: LintRuleReportFormatter


class SuppressedLintRuleReportFormatter(LintRuleReportFormatter):
    def _format_header(self, report: BaseLintRuleReport) -> str:
        return f"{super()._format_header(report)} [inserted suppression]"


def get_formatted_reports_for_path(
    path: Path,
    opts: InsertSuppressionsOpts,
    metadata_cache: Optional[Mapping["ProviderT", object]] = None,
) -> Iterable[str]:
    with open(path, "rb") as f:
        source = f.read()

    try:
        cst_wrapper = None
        if metadata_cache is not None:
            cst_wrapper = MetadataWrapper(
                parse_module(source),
                True,
                metadata_cache,
            )
        raw_reports = lint_file(
            path, source, rules={opts.rule}, cst_wrapper=cst_wrapper
        )
    except (SyntaxError, ParserSyntaxError) as e:
        print_red(
            f"Encountered the following error while parsing source code in file {path}:"
        )
        print(e)
        return []

    opts_message = opts.message
    comments = []
    for rr in raw_reports:
        if isinstance(opts_message, str):
            message = opts_message
        elif opts_message == MessageKind.USE_LINT_REPORT:
            message = rr.message
        else:  # opts_message == MessageKind.NO_MESSAGE
            message = None
        comments.append(
            SuppressionComment(opts.kind, rr.line, rr.code, message, opts.max_lines)
        )
    insert_suppressions_result = insert_suppressions(source, comments)
    updated_source = insert_suppressions_result.updated_source
    assert (
        not insert_suppressions_result.failed_insertions
    ), "Failed to insert some comments. This should not be possible."

    if updated_source != source:
        if not opts.skip_autoformatter:
            # Format the code using the config file's formatter.
            updated_source = invoke_formatter(
                get_lint_config().formatter, updated_source
            )
        with open(path, "wb") as f:
            f.write(updated_source)

    # linter completed successfully
    return [opts.formatter.format(rr) for rr in raw_reports]


def _add_arguments(parser: argparse.ArgumentParser) -> None:
    """All required arguments for `insert_supressions`"""
    parser.add_argument(
        "--kind",
        default="fixme",
        choices=[kind.name.lower() for kind in SuppressionCommentKind],
        help="Should we use `# lint-fixme` or `# lint-ignore`? Defaults to 'fixme'.",
    )
    message_group = parser.add_mutually_exclusive_group()
    message_group.add_argument(
        "--message",
        default=None,
        help="Overrides the lint message used in the fixme comment.",
    )
    message_group.add_argument(
        "--no-message",
        action="store_true",
        help=(
            "Don't include a message with the suppression comment. Only include the "
            + "lint code."
        ),
    )
    parser.add_argument(
        "--max-lines",
        default=3,
        type=int,
        help="The maximum number of lines a comment can span before getting truncated",
    )


def register_subparser(parser: Optional[argparse._SubParsersAction] = None) -> None:
    """Add parser or subparser for `insert_supressions` command."""
    if parser is None:
        insert_supressions_parser = argparse.ArgumentParser(
            description=DESCRIPTION, parents=PARENTS
        )
        _add_arguments(insert_supressions_parser)
        sys.exit(_main(insert_supressions_parser.parse_args()))

    else:
        insert_supressions_parser = parser.add_parser(
            "insert_suppressions",
            description=DESCRIPTION,
            parents=PARENTS,
            help="Insert comments where violations are found",
        )
        _add_arguments(insert_supressions_parser)
        insert_supressions_parser.set_defaults(subparser_fn=_main)


def _main(args: argparse.Namespace) -> int:
    width = shutil.get_terminal_size(fallback=(80, 24)).columns

    # Find files if directory was provided.
    file_paths = tuple(find_files((str(p) for p in args.paths)))

    if not args.compact:
        print(f"Scanning {len(file_paths)} files")
        print()
    start_time = time.time()

    if args.no_message:
        message = MessageKind.NO_MESSAGE
    elif args.message is not None:
        message = args.message
    else:
        message = MessageKind.USE_LINT_REPORT

    metadata_caches: Optional[Mapping[str, Mapping["ProviderT", object]]] = None
    if rules_require_metadata_cache({args.rule}):
        metadata_caches = get_metadata_caches(args.cache_timeout, file_paths)

    # opts is a more type-safe version of args that we pass around
    opts = InsertSuppressionsOpts(
        rule=args.rule,
        skip_autoformatter=args.skip_autoformatter,
        kind=SuppressionCommentKind[args.kind.upper()],
        message=message,
        max_lines=args.max_lines,
        formatter=SuppressedLintRuleReportFormatter(width, args.compact),
    )

    formatted_reports_iter = itertools.chain.from_iterable(
        map_paths(
            get_formatted_reports_for_path,
            file_paths,
            opts,
            workers=args.workers,
            metadata_caches=metadata_caches,
        )
    )

    formatted_reports = []
    for formatted_report in formatted_reports_iter:
        # Reports are yielded as soon as they're available. Stream the output to the
        # terminal.
        print(formatted_report)
        # save the report from the iterator for later use
        formatted_reports.append(formatted_report)

    if not args.compact:
        print()
        print(
            f"Found {len(formatted_reports)} reports in {len(file_paths)} files in "
            + f"{time.time() - start_time :.2f} seconds."
        )

    return 0


if __name__ == "__main__":
    register_subparser()



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from dataclasses import dataclass
from pathlib import Path
from typing import cast, Collection, List, Optional, Sequence

from libcst import Module
from libcst.testing.utils import UnitTest

from fixit.cli import LintOpts, map_paths
from fixit.cli.args import LintWorkers
from fixit.common.base import CstLintRule, LintConfig
from fixit.common.report import (
    BaseLintRuleReport,
    LintFailureReportBase,
    LintSuccessReportBase,
)
from fixit.rule_lint_engine import lint_file


@dataclass(frozen=True)
class FakeLintSuccessReport(LintSuccessReportBase):
    path: str
    status: str
    reports: Collection[str]

    @staticmethod
    # pyre-fixme[14]: `create_reports` overrides method defined in
    #  `LintSuccessReportBase` inconsistently.
    def create_reports(
        path: Path,
        reports: Collection[BaseLintRuleReport],
        global_list: List[str],
    ) -> Sequence["FakeLintSuccessReport"]:
        global_list.append(str(path))
        return [FakeLintSuccessReport(str(path), "success", ["fake picklable report"])]


@dataclass(frozen=True)
class FakeLintFailureReport(LintFailureReportBase):
    path: str
    status: str
    reports: Collection[str]

    @staticmethod
    # pyre-fixme[14]: `create_reports` overrides method defined in
    #  `LintFailureReportBase` inconsistently.
    def create_reports(
        path: Path,
        reports: Collection[BaseLintRuleReport],
        global_list: List[str],
    ) -> Sequence["FakeLintFailureReport"]:
        global_list.append(str(path))
        return [FakeLintFailureReport(str(path), "failure", ["fake picklable report"])]


class FakeRule(CstLintRule):
    def visit_Module(self, node: Module) -> None:
        self.report(node, "Dummy message")


def mock_operation(
    path: Path,
    opts: LintOpts,
    _=None,
) -> Sequence[FakeLintSuccessReport]:
    results = opts.success_report.create_reports(
        path,
        lint_file(path, b"test", rules=opts.rules, config=LintConfig()),
        **opts.extra,
    )
    return cast(Sequence[FakeLintSuccessReport], results)


def generate_mock_lint_opt(global_list: Optional[List[str]] = None) -> LintOpts:
    if global_list is None:
        global_list = []

    return LintOpts(
        {FakeRule},
        FakeLintSuccessReport,
        FakeLintFailureReport,
        extra={"global_list": global_list},
    )


class LintOptsTest(UnitTest):
    def setUp(self) -> None:
        self.global_list = []
        self.opts = generate_mock_lint_opt(global_list=self.global_list)

    def test_extra_opts(self) -> None:
        path = "path.py"
        results = next(
            map_paths(
                mock_operation,
                [path],
                self.opts,
                workers=LintWorkers.USE_CURRENT_THREAD,
            )
        )

        # Assert global list has been modified as expected.
        self.assertEqual(list(self.global_list), [path])

        # Assert the rest of the reporting functionality is as expected
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].reports, ["fake picklable report"])

    def test_extra_opts_multiple_paths(self) -> None:
        paths = ["path1.py", "path2.py"]
        results_iter = map_paths(
            mock_operation, paths, self.opts, workers=LintWorkers.USE_CURRENT_THREAD
        )
        results_count = 0
        paths_reported = []
        for results in results_iter:
            for result in results:
                results_count += 1
                paths_reported.append(result.path)
                self.assertEqual(len(result.reports), 1)
                self.assertEqual(result.reports, ["fake picklable report"])

        # Assert global list has been modified as expected.
        self.assertCountEqual(list(self.global_list), paths)

        # Assert all passed in paths have been visited as expected
        self.assertCountEqual(paths_reported, paths)



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import List, Union

import libcst as cst
import libcst.matchers as m

from fixit import (
    CstContext,
    CstLintRule,
    InvalidTestCase as Invalid,
    ValidTestCase as Valid,
)


class NoRedundantArgumentsSuperRule(CstLintRule):
    """
    Remove redundant arguments when using super for readability.
    """

    MESSAGE: str = (
        "Do not use arguments when calling super for the parent class. See "
        + "https://www.python.org/dev/peps/pep-3135/"
    )
    VALID = [
        Valid(
            """
            class Foo(Bar):
                def foo(self, bar):
                    super().foo(bar)
            """
        ),
        Valid(
            """
            class Foo(Bar):
                def foo(self, bar):
                    super(Bar, self).foo(bar)
            """
        ),
        Valid(
            """
            class Foo(Bar):
                @classmethod
                def foo(cls, bar):
                    super(Bar, cls).foo(bar)
            """
        ),
        Valid(
            """
            class Foo:
                class InnerBar(Bar):
                    def foo(self, bar):
                        pass

                class InnerFoo(InnerBar):
                    def foo(self, bar):
                        super(InnerBar, self).foo(bar)
            """
        ),
    ]
    INVALID = [
        Invalid(
            """
            class Foo(Bar):
                def foo(self, bar):
                    super(Foo, self).foo(bar)
            """,
            expected_replacement="""
            class Foo(Bar):
                def foo(self, bar):
                    super().foo(bar)
            """,
        ),
        Invalid(
            """
            class Foo(Bar):
                @classmethod
                def foo(cls, bar):
                    super(Foo, cls).foo(bar)
            """,
            expected_replacement="""
            class Foo(Bar):
                @classmethod
                def foo(cls, bar):
                    super().foo(bar)
            """,
        ),
        Invalid(
            """
            class Foo:
                class InnerFoo(Bar):
                    def foo(self, bar):
                        super(Foo.InnerFoo, self).foo(bar)
            """,
            expected_replacement="""
            class Foo:
                class InnerFoo(Bar):
                    def foo(self, bar):
                        super().foo(bar)
            """,
        ),
        Invalid(
            """
            class Foo:
                class InnerFoo(Bar):
                    class InnerInnerFoo(Bar):
                        def foo(self, bar):
                            super(Foo.InnerFoo.InnerInnerFoo, self).foo(bar)
            """,
            expected_replacement="""
            class Foo:
                class InnerFoo(Bar):
                    class InnerInnerFoo(Bar):
                        def foo(self, bar):
                            super().foo(bar)
            """,
        ),
    ]

    def __init__(self, context: CstContext) -> None:
        super().__init__(context)
        self.current_classes: List[str] = []

    def visit_ClassDef(self, node: cst.ClassDef) -> None:
        self.current_classes.append(node.name.value)

    def leave_ClassDef(self, original_node: cst.ClassDef) -> None:
        self.current_classes.pop()

    def leave_Call(self, original_node: cst.Call) -> None:
        if self.current_classes and m.matches(
            original_node,
            m.Call(
                func=m.Name("super"),
                args=[
                    m.Arg(value=self._build_arg_class_matcher()),
                    m.Arg(),
                ],
            ),
        ):
            self.report(original_node, replacement=original_node.with_changes(args=()))

    def _build_arg_class_matcher(self) -> Union[m.Attribute, m.Name]:
        matcher = m.Name(value=self.current_classes[0])

        # For nested classes, we need to match attributes, so we can target
        # `super(Foo.InnerFoo, self)` for example.
        if len(self.current_classes) > 1:
            for class_name in self.current_classes[1:]:
                matcher = m.Attribute(value=matcher, attr=m.Name(value=class_name))

        return matcher



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


def print_color(code: int, message: str) -> None:
    print(f"\033[{code}m{message}\033[00m")


def print_green(message: str) -> None:
    print_color(92, message)


def print_yellow(message: str) -> None:
    print_color(93, message)


def print_cyan(message: str) -> None:
    print_color(96, message)


def print_red(message: str) -> None:
    print_color(91, message)


def snake_to_camelcase(name: str) -> str:
    """Convert snake-case string to camel-case string."""
    return "".join(n.capitalize() for n in name.split("_"))



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
Shared utilities for tools that need to run lint rules from the command line.
"""
import argparse
import itertools
import json
import multiprocessing
import os
import traceback
import warnings
from dataclasses import asdict, dataclass, field
from pathlib import Path
from typing import (
    Callable,
    Collection,
    Dict,
    Iterable,
    Iterator,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Type,
    TYPE_CHECKING,
    TypeVar,
    Union,
)

import libcst as cst
from libcst.metadata import MetadataWrapper

from fixit.cli.args import get_multiprocessing_parser, LintWorkers
from fixit.common.base import LintConfig
from fixit.common.config import get_lint_config
from fixit.common.full_repo_metadata import FullRepoMetadataConfig, get_repo_caches
from fixit.common.report import LintFailureReportBase, LintSuccessReportBase
from fixit.common.utils import LintRuleCollectionT
from fixit.rule_lint_engine import lint_file


if TYPE_CHECKING:
    from libcst.metadata.base_provider import ProviderT

_MapPathsOperationConfigT = TypeVar("_MapPathsOperationConfigT")
_MapPathsOperationResultT = TypeVar("_MapPathsOperationResultT")
_MapPathsOperationT = Callable[
    [Path, _MapPathsOperationConfigT, Optional[Mapping["ProviderT", object]]],
    _MapPathsOperationResultT,
]
_MapPathsWorkerArgsT = Tuple[
    _MapPathsOperationT,
    str,
    _MapPathsOperationConfigT,
    Optional[Mapping["ProviderT", object]],
]


def find_files(paths: Iterable[Union[str, Path]]) -> Iterator[str]:
    """
    Given an iterable of paths, yields any files and walks over any directories.
    """
    for path in paths:
        if os.path.isfile(path):
            yield str(path)
        else:
            for root, _dirs, files in os.walk(path):
                for f in files:
                    if f.endswith(".py") and not os.path.islink(f):
                        yield os.path.join(root, f)


# Multiprocessing can only pass one argument. Wrap `operation` to provide this.
# pyre-fixme[34]: `Variable[_MapPathsOperationResultT]` isn't present in the
#  function's parameters.
def _map_paths_worker(args: _MapPathsWorkerArgsT) -> _MapPathsOperationResultT:
    operation, path, config, metadata_caches = args
    return operation(Path(path), config, metadata_caches)


def map_paths(
    operation: _MapPathsOperationT,
    paths: Iterable[str],
    config: _MapPathsOperationConfigT,
    *,
    workers: Union[int, LintWorkers] = LintWorkers.CPU_COUNT,
    metadata_caches: Optional[Mapping[str, Mapping["ProviderT", object]]] = None,
    # pyre-fixme[34]: `Variable[_MapPathsOperationResultT]` isn't present in the
    #  function's parameters.
) -> Iterator[_MapPathsOperationResultT]:
    """
    Applies the given `operation` to each file path in `paths`.

    This uses a process pool by default, but if `workers` is
    `LintWorkers.USE_CURRENT_THREAD`, the process pool is disabled and rules are
    processed on the current thread. This is useful for profiling and debugging.

    `operation` must be a top-level function (not a method or inner function), since it
    needs to be imported by pickle and used across process boundaries.
    NOTE: this function does not verify the signature of `operation`. If `metadata_caches`
    is passed in, it is up to the caller to make sure that `operation` is equipped to handle
    a `MetadataWrapper` argument.

    `paths` should only contain file paths (not directories). Use `find_files` if you
    have directory paths and need to expand them.

    `metadata_caches` is an optional argument for those callers expecting to use type metadata
    for linting. If passed, it should be a mapping of Path to a LibCST MetadataWrapper to be used
    for the linting of the corresponding file.

    Results are yielded as soon as they're available, so they may appear out-of-order.
    """
    if workers is LintWorkers.CPU_COUNT:
        workers = multiprocessing.cpu_count()

    if metadata_caches is not None:
        tasks: Collection[_MapPathsWorkerArgsT] = tuple(
            zip(
                itertools.repeat(operation),
                metadata_caches.keys(),
                itertools.repeat(config),
                metadata_caches.values(),
            )
        )
    else:
        tasks: Collection[_MapPathsWorkerArgsT] = tuple(
            zip(
                itertools.repeat(operation),
                paths,
                itertools.repeat(config),
                itertools.repeat(None),
            )
        )
    if not tasks:
        # this would result in 0 workers, which will cause multiprocessing.Pool to die
        return

    if workers is LintWorkers.USE_CURRENT_THREAD:
        for t in tasks:
            yield _map_paths_worker(t)
    else:
        assert not isinstance(workers, LintWorkers), "Unreachable"
        # Don't spawn more processes than there are tasks. Multiprocessing is eager and
        # will spawn workers immediately even if there's no work for them to do.
        with multiprocessing.Pool(min(workers, len(tasks))) as pool:
            for result in pool.imap_unordered(_map_paths_worker, tasks):
                yield result


@dataclass(frozen=True)
class LintOpts:
    rules: LintRuleCollectionT
    success_report: Type[LintSuccessReportBase]
    failure_report: Type[LintFailureReportBase]
    config: LintConfig = get_lint_config()
    full_repo_metadata_config: Optional[FullRepoMetadataConfig] = None
    extra: Dict[str, object] = field(default_factory=dict)


def get_file_lint_result_json(
    path: Path,
    opts: LintOpts,
    metadata_cache: Optional[Mapping["ProviderT", object]] = None,
) -> Sequence[str]:
    try:
        with open(path, "rb") as f:
            source = f.read()
        cst_wrapper = None
        if metadata_cache is not None:
            cst_wrapper = MetadataWrapper(
                cst.parse_module(source),
                True,
                metadata_cache,
            )
        results = opts.success_report.create_reports(
            path,
            lint_file(
                path,
                source,
                rules=opts.rules,
                config=opts.config,
                cst_wrapper=cst_wrapper,
                find_unused_suppressions=True,
            ),
            **opts.extra,
        )
    except Exception:
        tb_str = traceback.format_exc()
        results = opts.failure_report.create_reports(path, tb_str, **opts.extra)
    return [json.dumps(asdict(r)) for r in results]


@dataclass(frozen=True)
class IPCResult:
    paths: List[str]


def run_ipc(
    opts: LintOpts,
    paths: List[str],
    prefix: Optional[str] = None,
    workers: LintWorkers = LintWorkers.CPU_COUNT,
) -> IPCResult:
    """
    Given a LintOpts config with lint rules and lint success/failure report formatter,
    this IPC helper takes a path of source files (with an optional `prefix` that will be prepended).
    Results are formed as JSON and delimited by newlines.
    It uses a multiprocess pool and the results are streamed to stdout as soon
    as they're available.

    Returns an IPCResult object.
    """

    resolved_paths = (os.path.join(prefix, p) if prefix else p for p in paths)

    full_repo_metadata_config = opts.full_repo_metadata_config
    metadata_caches: Optional[Mapping[str, Mapping["ProviderT", object]]] = None
    if full_repo_metadata_config is not None:
        metadata_caches = get_repo_caches(resolved_paths, full_repo_metadata_config)

    results_iter: Iterator[Sequence[str]] = map_paths(
        get_file_lint_result_json,
        resolved_paths,
        opts,
        workers=workers,
        metadata_caches=metadata_caches,
    )
    for results in results_iter:
        # Use print outside of the executor to avoid multiple processes trying to write
        # to stdout in parallel, which could cause a corrupted output.
        for result in results:
            print(result)

    return IPCResult(list(resolved_paths))


def ipc_main(opts: LintOpts) -> IPCResult:
    """
    Like `run_ipc` instead this function expects arguments to be collected through
    argparse. This IPC helper takes paths of source files from either a path file
    (with @paths arg) or a list of paths as args.

    Returns an IPCResult object.
    """
    warnings.warn(
        """
        Calling ipc_main as a command line tool is being deprecated.
        Please use the module-level function `run_ipc` instead.""",
        DeprecationWarning,
    )

    parser = argparse.ArgumentParser(
        description="Runs Fixit lint rules and print results as console output.",
        fromfile_prefix_chars="@",
        parents=[get_multiprocessing_parser()],
    )
    parser.add_argument("paths", nargs="*", help="List of paths to run lint rules on.")
    parser.add_argument("--prefix", help="A prefix to be added to all paths.")
    args: argparse.Namespace = parser.parse_args()

    return run_ipc(
        opts=opts, paths=args.paths, prefix=args.prefix, workers=args.workers
    )



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# Usage:
#
#   $ python -m fixit.cli.run_rules --help
#   $ python -m fixit.cli.run_rules
#   $ python -m fixit.cli.run_rules --rules AvoidOrInExceptRule
#   $ python -m fixit.cli.run_rules . --rules AvoidOrInExceptRule NoUnnecessaryListComprehensionRule
#   $ python -m fixit.cli.run_rules . --rules AvoidOrInExceptRule my.custom.rules.package
#   $ python -m fixit.cli.run_rules . --rules fixit.rules

import argparse
import itertools
import shutil
import sys
import time
from dataclasses import dataclass
from pathlib import Path
from typing import Iterable, List, Mapping, Optional, TYPE_CHECKING

from libcst import parse_module, ParserSyntaxError
from libcst.metadata import MetadataWrapper

from fixit.cli import find_files, map_paths
from fixit.cli.args import (
    get_compact_parser,
    get_metadata_cache_parser,
    get_multiprocessing_parser,
    get_paths_parser,
    get_rules_parser,
    get_skip_ignore_byte_marker_parser,
    get_skip_ignore_comments_parser,
)
from fixit.cli.formatter import LintRuleReportFormatter
from fixit.cli.full_repo_metadata import (
    get_metadata_caches,
    rules_require_metadata_cache,
)
from fixit.cli.utils import print_red
from fixit.common.utils import LintRuleCollectionT
from fixit.rule_lint_engine import lint_file


if TYPE_CHECKING:
    from libcst.metadata.base_provider import ProviderT

DESCRIPTION: str = """Validates your lint rules by running them against the specified,
directory or file(s). This is not a substitute for unit tests, but it can provide
additional confidence in your lint rules. If no lint rules or packages are specified,
runs all lint rules found in the packages specified in `fixit.config.yaml`."""

PARENTS: List[argparse.ArgumentParser] = [
    get_paths_parser(),
    get_rules_parser(),
    get_skip_ignore_byte_marker_parser(),
    get_skip_ignore_comments_parser(),
    get_compact_parser(),
    get_multiprocessing_parser(),
    get_metadata_cache_parser(),
]


@dataclass(frozen=True)
class LintOpts:
    rules: LintRuleCollectionT
    use_ignore_byte_markers: bool
    use_ignore_comments: bool
    formatter: LintRuleReportFormatter


def get_formatted_reports_for_path(
    path: Path,
    opts: LintOpts,
    metadata_cache: Optional[Mapping["ProviderT", object]] = None,
) -> Iterable[str]:
    with open(path, "rb") as f:
        source = f.read()

    try:
        cst_wrapper = None
        if metadata_cache is not None:
            cst_wrapper = MetadataWrapper(parse_module(source), True, metadata_cache)
        raw_reports = lint_file(
            path,
            source,
            rules=opts.rules,
            use_ignore_byte_markers=opts.use_ignore_byte_markers,
            use_ignore_comments=opts.use_ignore_comments,
            cst_wrapper=cst_wrapper,
            find_unused_suppressions=True,
        )
    except (SyntaxError, ParserSyntaxError) as e:
        print_red(
            f"Encountered the following error while parsing source code in file {path}:"
        )
        print(e)
        return []

    # linter completed successfully
    return [opts.formatter.format(rr) for rr in raw_reports]


def _add_arguments(parser: argparse.ArgumentParser) -> None:
    """All required arguments for `run_rules`"""
    pass


def register_subparser(parser: Optional[argparse._SubParsersAction] = None) -> None:
    """Add parser or subparser for `run_rules` command."""
    if parser is None:
        run_rules_parser = argparse.ArgumentParser(
            description=DESCRIPTION,
            parents=PARENTS,
        )
        _add_arguments(run_rules_parser)
        sys.exit(_main(run_rules_parser.parse_args()))

    else:
        run_rules_parser = parser.add_parser(
            "run_rules",
            description=DESCRIPTION,
            parents=PARENTS,
            help="Run fixit rules against python code",
        )
        _add_arguments(run_rules_parser)
        run_rules_parser.set_defaults(subparser_fn=_main)


def _main(args: argparse.Namespace) -> int:
    width = shutil.get_terminal_size(fallback=(80, 24)).columns

    # expand path if it's a directory
    file_paths = tuple(find_files(args.paths))
    all_rules = args.rules

    if not args.compact:
        print(f"Scanning {len(file_paths)} files")
        print(f"Testing {len(all_rules)} rules")
        print()
    start_time = time.time()

    metadata_caches: Optional[Mapping[str, Mapping["ProviderT", object]]] = None
    if rules_require_metadata_cache(all_rules):
        metadata_caches = get_metadata_caches(args.cache_timeout, file_paths)

    # opts is a more type-safe version of args that we pass around
    opts = LintOpts(
        rules=all_rules,
        use_ignore_byte_markers=args.use_ignore_byte_markers,
        use_ignore_comments=args.use_ignore_comments,
        formatter=LintRuleReportFormatter(width, args.compact),
    )

    formatted_reports_iter = itertools.chain.from_iterable(
        map_paths(
            get_formatted_reports_for_path,
            file_paths,
            opts,
            workers=args.workers,
            metadata_caches=metadata_caches,
        )
    )

    formatted_reports = []
    for formatted_report in formatted_reports_iter:
        # Reports are yielded as soon as they're available. Stream the output to the
        # terminal.
        print(formatted_report)
        # save the report from the iterator for later use
        formatted_reports.append(formatted_report)

    if not args.compact:
        print()
        print(
            f"Found {len(formatted_reports)} reports in {len(file_paths)} files in "
            + f"{time.time() - start_time :.2f} seconds."
        )

    # Return with an exit code of 1 if there are any violations found.
    return int(bool(formatted_reports))


if __name__ == "__main__":
    register_subparser()
